# coding=utf-8
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for android_env.components.utils."""

from absl.testing import absltest
from absl.testing import parameterized
from android_env.components import utils
from android_env.proto import adb_pb2
import numpy as np


class UtilsTest(parameterized.TestCase):

  @parameterized.parameters(
      ([0.5, 0.5], [320, 480], (160, 240)),
      ([0.25, 0.75], [320, 480], (80, 360)),
      ([0.0, 0.0], [320, 480], (0, 0)),
      ([1.0, 1.0], [320, 480], (319, 479)),
      )
  def test_touch_position_to_pixel_position(
      self, touch_pos, width_height, pixel_pos):
    self.assertEqual(utils.touch_position_to_pixel_position(
        np.array(touch_pos), width_height), pixel_pos)

  def test_transpose_pixels(self):
    image = np.reshape(np.array(range(12)), (3, 2, 2))
    expected = [[[0, 1], [4, 5], [8, 9]], [[2, 3], [6, 7], [10, 11]]]
    self.assertEqual(utils.transpose_pixels(image).shape, (2, 3, 2))
    self.assertTrue((utils.transpose_pixels(image) == expected).all())

  def test_orient_pixels(self):
    image = np.reshape(np.array(range(12)), (3, 2, 2))

    expected_90 = [[[8, 9], [4, 5], [0, 1]], [[10, 11], [6, 7], [2, 3]]]
    rot_90 = adb_pb2.AdbCall.Rotate.Orientation.LANDSCAPE_90
    rotated = utils.orient_pixels(image, rot_90)
    self.assertEqual(rotated.shape, (2, 3, 2))
    self.assertTrue((rotated == expected_90).all())

    expected_180 = [[[10, 11], [8, 9]], [[6, 7], [4, 5]], [[2, 3], [0, 1]]]
    rot_180 = adb_pb2.AdbCall.Rotate.Orientation.PORTRAIT_180
    rotated = utils.orient_pixels(image, rot_180)
    self.assertEqual(rotated.shape, (3, 2, 2))
    self.assertTrue((rotated == expected_180).all())

    expected_270 = [[[2, 3], [6, 7], [10, 11]], [[0, 1], [4, 5], [8, 9]]]
    rot_270 = adb_pb2.AdbCall.Rotate.Orientation.LANDSCAPE_270
    rotated = utils.orient_pixels(image, rot_270)
    self.assertEqual(rotated.shape, (2, 3, 2))
    self.assertTrue((rotated == expected_270).all())

    rot_0 = adb_pb2.AdbCall.Rotate.Orientation.PORTRAIT_0
    rotated = utils.orient_pixels(image, rot_0)
    self.assertEqual(rotated.shape, (3, 2, 2))
    self.assertTrue((rotated == image).all())

if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# vim: set tabstop=2 shiftwidth=2:
# Copyright 2023 SJTU X-Lance Lab
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Revised by Danyang Zhang @X-Lance based on
# 
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wraps the AndroidEnv environment to provide discrete actions."""

from typing import Optional, Sequence
from typing import Dict, Tuple

import android_env
from android_env.components import action_type
from android_env.wrappers import base_wrapper
import dm_env
from dm_env import specs
import numpy as np

NOISE_CLIP_VALUE = 0.4999

class DiscreteActionWrapper(base_wrapper.BaseWrapper):
  """AndroidEnv with discrete actions."""

  def __init__(self,
               env: dm_env.Environment,
               action_grid: Sequence[int] = (10, 10),
               redundant_actions: bool = False,
               keep_repeat: bool = False,
               noise: float = 0.1):
    #  method `__init__` {{{ # 
    """
    Args:
        env: android_env.AndroidEnv
        action_grid: Sequence of int as (height, width)
        redundant_actions: bool
          - if True, then the action space will be ((G+V)*A,) where G is grid
            size, V is vocabulary size and A is the number of action types;
            action type ids are assigned to actions TOUCH, LIFT, REPEAT, TEXT
            by sequence where REPEAT is optional and is controlled by
            `keep_repeat` and TEXT is optional and is determined by the
            vocabulary size of the internal `env`
          - if False, the action space will be flatten to cat(TO, L, R, TE);
            the length of the TOUCH segment is G; LIFT and REPEAT occupy 1 slot
            respectively; TEXT segment is V-long. REPEAT and TEXT are also
            optional and the condition is as well as the case that
            `redundant_actions`==True
        keep_repeat: bool
        noise: floating
    """

    super().__init__(env)

    self._action_grid: Tuple[int, int] = tuple(action_grid)  # [height, width]
    self._grid_size: np.int64 = np.prod(self._action_grid)

    self._redundant_actions: bool = redundant_actions
    self._keep_repeat = keep_repeat
    self._noise: float = noise

    self._post_switch_task()
    #  }}} method `__init__` # 

  def _assert_base_env(self):
    """Checks that the wrapped env has the right action spec format."""

    assert len(self._parent_action_spec) == 2 or len(self._parent_action_spec) == 3
    assert not self._parent_action_spec['action_type'].shape
    assert self._parent_action_spec['touch_position'].shape == (2,)
    assert not self._parent_action_spec['input_token'].shape if "input_token" in self._parent_action_spec else True

  @property
  def num_actions(self) -> int:
    """Number of discrete actions."""

    if self._redundant_actions:
      return (self._grid_size+self._vocabulary_size) * self._num_action_types
    else:
      return self._grid_size + self._num_action_types - 1 + max(0, self._vocabulary_size-1)

  def step(self, action: Dict[str, int]) -> dm_env.TimeStep:
    """Take a step in the base environment."""

    return self._env.step(self._process_action(action))

  def _process_action(self, action: Dict[str, int]) -> Dict[str, np.ndarray]:
    """Transforms action so that it agrees with AndroidEnv's action spec."""

    action_type_ = self._get_action_type(action['action_id'])
    if action_type_!=action_type.ActionType.TEXT:
      return {
          'action_type':
              np.array(action_type_,
                       dtype=self._parent_action_spec['action_type'].dtype),
          'touch_position':
              np.array(self._get_touch_position(action['action_id']),
                       dtype=self._parent_action_spec['touch_position'].dtype)
      }
    else:
      return {
          "action_type":
            np.array(action_type_,
              dtype=self._parent_action_spec["action_type"].dtype),
          "input_token":
            np.array(action["action_id"]-self._grid_size-self._num_action_types+2,
              dtype=self._parent_action_spec["input_token"].dtype)
        }

  def _get_action_type(self, action_id: int) -> action_type.ActionType:
    #  method `_get_action_type` {{{ # 
    """Compute action type corresponding to the given action_id.

    When `self._redundant_actions` == True the `grid_size` is "broadcast" over
    all the possible actions so you end up with `grid_size` discrete actions
    of type 0, `grid_size` discrete actions of type 1, etc. for all action
    types.

    When `self._redundant_actions` == False the first `grid_size` actions are
    reserved for "touch" and the rest are just added (NOT multiplied) to the
    total number of discrete actions (exactly one of LIFT and REPEAT).

    Args:
      action_id: A discrete action.
    Returns:
      action_type: The action_type of the action.
    """

    if self._redundant_actions:
      assert action_id < self.num_actions
      return action_id // (self._grid_size+self._vocabulary_size)

    else:
      assert action_id <= self.num_actions
      if action_id < self._grid_size:
        return action_type.ActionType.TOUCH
      elif action_id == self._grid_size:
        return action_type.ActionType.LIFT
      else:
        if self._keep_repeat:
          return action_type.ActionType.REPEAT if action_id==self._grid_size+1\
              else action_type.ActionType.TEXT
        else:
          return action_type.ActionType.TEXT
    #  }}} method `_get_action_type` # 

  def _get_touch_position(self, action_id: int) -> Sequence[float]:
    #  method `_get_touch_position` {{{ # 
    """Compute the position corresponding to the given action_id.

    Note: in the touch_position (x, y) of an action, x corresponds to the
    horizontal axis (width), and y corresponds to the vertical axis (height)
    of the screen. BUT, the screen has dimensions (height, width), i.e. the
    first coordinate corresponds to y, and the second coordinate corresponds
    to x. Pay attention to this mismatch in the calculations below.

    Args:
      action_id: A discrete action.
    Returns:
      touch_position: The [0,1]x[0,1] coordinate of the action.
    """

    position_idx = action_id % self._grid_size

    x_pos_grid = position_idx % self._action_grid[1]  # WIDTH
    y_pos_grid = position_idx // self._action_grid[1]  # HEIGHT

    noise_x = np.random.normal(loc=0.0, scale=self._noise)
    noise_y = np.random.normal(loc=0.0, scale=self._noise)

    # Noise is clipped so that the action will strictly stay in the cell.
    noise_x = max(min(noise_x, NOISE_CLIP_VALUE), -NOISE_CLIP_VALUE)
    noise_y = max(min(noise_y, NOISE_CLIP_VALUE), -NOISE_CLIP_VALUE)

    x_pos = (x_pos_grid + 0.5 + noise_x) / self._action_grid[1]  # WIDTH
    y_pos = (y_pos_grid + 0.5 + noise_y) / self._action_grid[0]  # HEIGHT

    # Project action space to action_spec ranges. For the default case of
    # minimum = [0, 0] and maximum = [1, 1], this will not do anything.
    x_min, y_min = self._parent_action_spec['touch_position'].minimum
    x_max, y_max = self._parent_action_spec['touch_position'].maximum

    x_pos = x_min + x_pos * (x_max - x_min)
    y_pos = y_min + y_pos * (y_max - y_min)

    return [x_pos, y_pos]
    #  }}} method `_get_touch_position` # 

  def action_spec(self) -> Dict[str, specs.Array]:
    """Action spec of the wrapped environment."""

    return {
        'action_id':
            specs.DiscreteArray(
                num_values=self.num_actions,
                name='action_id')
    }

  def _post_switch_task(self):
    self._parent_action_spec: Dict[str, specs.Array] = self._env.action_spec()
    self._assert_base_env()

    self._num_action_types: int = self._parent_action_spec['action_type'].num_values
    if not self._keep_repeat:
      self._num_action_types -= 1
    self._vocabulary_size: int = self._parent_action_spec['input_token'].num_values\
            if "input_token" in self._parent_action_spec else 0



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: android_env/proto/raw_observation.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\'android_env/proto/raw_observation.proto\x12\x0b\x61ndroid_env\"\xa5\x01\n\x0eRawObservation\x12\x14\n\x0ctimestamp_us\x18\x01 \x01(\x03\x12\x31\n\x06screen\x18\x02 \x01(\x0b\x32!.android_env.RawObservation.Image\x1aJ\n\x05Image\x12\r\n\x05width\x18\x01 \x01(\x05\x12\x0e\n\x06height\x18\x02 \x01(\x05\x12\x14\n\x0cnum_channels\x18\x03 \x01(\x05\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\x0c\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'android_env.proto.raw_observation_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _RAWOBSERVATION._serialized_start=57
  _RAWOBSERVATION._serialized_end=222
  _RAWOBSERVATION_IMAGE._serialized_start=148
  _RAWOBSERVATION_IMAGE._serialized_end=222
# @@protoc_insertion_point(module_scope)



# coding=utf-8
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for android_env.components.app_screen_checker."""

import re
from typing import Sequence

from absl.testing import absltest
from android_env.components import app_screen_checker


def flatten_tree(tree: app_screen_checker.DumpsysNode,
                 flat_tree: Sequence[str],
                 indent: int = 2):
  """Appends a list of strings to `flat_tree` from `tree`."""
  flat_tree.append(' ' * indent + tree.data)
  for c in tree.children:
    flatten_tree(c, flat_tree, indent + 2)


class AppScreenCheckerTest(absltest.TestCase):

  # Ensures that build_tree_from_dumpsys_output produces a node whose flat
  # representation matches our expectation from an arbitrary hierarchy.
  def test_build_tree_from_dumpsys_output(self):
    dumpsys_output = """
Queen Elizabeth II
  Charles
    William
      George
      Charlotte
      Louis
    Harry
      Archie
  Anne
    Peter
      Savannah
      Isla
    Zara
      Mia
      Lena
  Andrew
    Beatrice
    Eugenie
  Edward
    Louise
    James
"""
    tree = app_screen_checker.build_tree_from_dumpsys_output(dumpsys_output)
    flat_tree = []
    flatten_tree(tree, flat_tree, indent=2)
    self.assertEqual(flat_tree, [
        '  ___root___',
        '    Queen Elizabeth II',
        '      Charles',
        '        William',
        '          George',
        '          Charlotte',
        '          Louis',
        '        Harry',
        '          Archie',
        '      Anne',
        '        Peter',
        '          Savannah',
        '          Isla',
        '        Zara',
        '          Mia',
        '          Lena',
        '      Andrew',
        '        Beatrice',
        '        Eugenie',
        '      Edward',
        '        Louise',
        '        James',
    ])

  # Ensures that build_tree_from_dumpsys_output produces a node whose flat
  # representation matches our expectation from an arbitrary hierarchy.
  def test_build_forest_from_dumpsys_output(self):
    dumpsys_output = """
Tree1
  Branch1
    Leaf1
    Leaf2
  Branch2
    Leaf3
    Leaf4
    Leaf5
Tree2
  Branch3
    Leaf6
    Leaf7
  Branch4
    Leaf8
    Leaf9
    Leaf10
  Leaf11
"""
    tree = app_screen_checker.build_tree_from_dumpsys_output(dumpsys_output)
    flat_tree = []
    flatten_tree(tree, flat_tree, indent=2)
    self.assertEqual(flat_tree, [
        '  ___root___',
        '    Tree1',
        '      Branch1',
        '        Leaf1',
        '        Leaf2',
        '      Branch2',
        '        Leaf3',
        '        Leaf4',
        '        Leaf5',
        '    Tree2',
        '      Branch3',
        '        Leaf6',
        '        Leaf7',
        '      Branch4',
        '        Leaf8',
        '        Leaf9',
        '        Leaf10',
        '      Leaf11',
    ])

  def test_no_view_hierarchy_matches_path(self):
    dumpsys_output = """
TASK
  ACTIVITY
    Missing View Hierarchy
      A
        B
        C
      D
        E
          F
"""
    expected_path = ['^A$', 'B$']
    expected_view_hierarchy_path = [
        re.compile(regex) for regex in expected_path
    ]
    self.assertFalse(
        app_screen_checker.matches_path(dumpsys_output,
                                        expected_view_hierarchy_path))

  def test_matches_path(self):
    dumpsys_output = """
TASK
  ACTIVITY
    Some node we don't care
      Blah

    View Hierarchy
      Hirohito
        Akihito
          Naruhito
            Aiko
          Fumihito
            Mako
            Kako
            Hisahito
        Masahito
"""
    expected_path = ['^Hirohito$', 'Akihito$', 'Fumihito$', 'Kako$']
    expected_view_hierarchy_path = [
        re.compile(regex) for regex in expected_path
    ]
    self.assertTrue(
        app_screen_checker.matches_path(
            dumpsys_output, expected_view_hierarchy_path, max_levels=2))

    # Also check that the following path does not match anything in the tree.
    expected_path = ['^Hirohito$', 'Akihito$', 'Fumihito$', 'Kenji$']
    expected_view_hierarchy_path = [
        re.compile(regex) for regex in expected_path
    ]
    self.assertFalse(
        app_screen_checker.matches_path(dumpsys_output,
                                        expected_view_hierarchy_path))

  def test_matches_path_one_level_deep(self):
    dumpsys_output = """
TASK
  ACTIVITY
    Some node we don't care
      Blah

    Some intermediate node
      View Hierarchy
        Hirohito
          Akihito
            Naruhito
              Aiko
            Fumihito
              Mako
              Kako
              Hisahito
          Masahito
"""
    expected_path = ['^Hirohito$', 'Akihito$', 'Fumihito$', 'Kako$']
    expected_view_hierarchy_path = [
        re.compile(regex) for regex in expected_path
    ]
    self.assertTrue(
        app_screen_checker.matches_path(
            dumpsys_output, expected_view_hierarchy_path, max_levels=3))

    # Also check that the view hierarchy is not found when searching only grand
    # children of TASK.
    expected_path = ['^Hirohito$', 'Akihito$', 'Fumihito$', 'Kako$']
    expected_view_hierarchy_path = [
        re.compile(regex) for regex in expected_path
    ]
    self.assertFalse(
        app_screen_checker.matches_path(
            dumpsys_output, expected_view_hierarchy_path, max_levels=2))


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for android_env.components.task_manager.py."""

import json

from absl.testing import absltest
from android_env.components import adb_controller
from android_env.components import dumpsys_thread
from android_env.components import log_stream
from android_env.components import logcat_thread
from android_env.components import setup_step_interpreter
from android_env.components import task_manager
from android_env.proto import task_pb2
import mock
import numpy as np


class TaskManagerTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.addCleanup(mock.patch.stopall)  # Disable previous patches.

    self._adb_controller = mock.create_autospec(adb_controller.AdbController)
    self._setup_step_interpreter = mock.create_autospec(
        setup_step_interpreter.SetupStepInterpreter)
    self._dumpsys_thread = mock.create_autospec(dumpsys_thread.DumpsysThread)
    self._logcat_thread = mock.create_autospec(logcat_thread.LogcatThread)
    self._log_stream = mock.create_autospec(log_stream.LogStream)

    mock.patch.object(
        adb_controller, 'AdbController',
        return_value=self._adb_controller).start()
    mock.patch.object(
        setup_step_interpreter,
        'SetupStepInterpreter',
        return_value=self._setup_step_interpreter).start()
    mock.patch.object(
        dumpsys_thread, 'DumpsysThread',
        return_value=self._dumpsys_thread).start()
    mock.patch.object(
        logcat_thread, 'LogcatThread',
        return_value=self._logcat_thread).start()
    mock.patch.object(
        log_stream, 'LogStream',
        return_value=self._log_stream).start()

  def test_setup_task(self):
    task_mgr = task_manager.TaskManager(task=task_pb2.Task())
    task_mgr.setup_task(
        adb_controller=self._adb_controller,
        log_stream=self._log_stream)
    assert hasattr(task_mgr, '_logcat_thread')
    assert hasattr(task_mgr, '_setup_step_interpreter')

  def test_get_current_reward(self):
    # Replace `LogcatThread.add_event_listener` with one that simply calls `fn`
    # right away.
    def my_add_ev_listener(event_listener: logcat_thread.EventListener):
      # Check that the event matches what's expected.
      match = event_listener.regexp.match('Reward: 123.0')
      if match is None:  # Ignore events that are not rewards.
        return

      event_listener.handler_fn(event_listener.regexp, match)

    task = task_pb2.Task()
    task.log_parsing_config.log_regexps.reward.extend([
        '^[Rr]eward: ([-+]?[0-9]*\\.?[0-9]*)$'
    ])
    task_mgr = task_manager.TaskManager(task=task)
    self._logcat_thread.add_event_listener.side_effect = my_add_ev_listener
    task_mgr.setup_task(
        adb_controller=self._adb_controller,
        log_stream=self._log_stream)
    self.assertEqual(task_mgr.get_current_reward(), 123.0)

  def test_reward_event(self):
    # Replace `LogcatThread.add_event_listener` with one that simply calls `fn`
    # right away.
    def my_add_ev_listener(event_listener: logcat_thread.EventListener):
      # Check that the event matches what's expected.
      match_1 = event_listener.regexp.match('foo_1')
      match_2 = event_listener.regexp.match('foo_2')
      match_3 = event_listener.regexp.match('Reward: 2.0')
      if match_1:
        event_listener.handler_fn(event_listener.regexp, match_1)
      if match_2:
        event_listener.handler_fn(event_listener.regexp, match_2)
      if match_3:
        event_listener.handler_fn(event_listener.regexp, match_3)

    task = task_pb2.Task()
    reward_event_1 = task_pb2.LogParsingConfig.LogRegexps.RewardEvent(
        event='foo_1', reward=5.0)
    reward_event_2 = task_pb2.LogParsingConfig.LogRegexps.RewardEvent(
        event='foo_2', reward=-1.0)
    task.log_parsing_config.log_regexps.reward_event.extend(
        [reward_event_1, reward_event_2])
    task.log_parsing_config.log_regexps.reward.extend(
        ['^[Rr]eward: ([-+]?[0-9]*\\.?[0-9]*)$'])
    task_mgr = task_manager.TaskManager(task=task)
    self._logcat_thread.add_event_listener.side_effect = my_add_ev_listener
    task_mgr.setup_task(
        adb_controller=self._adb_controller,
        log_stream=self._log_stream)
    self.assertEqual(task_mgr.get_current_reward(), 6.0)

  def test_get_current_reward_via_score(self):
    # Replace `LogcatThread.add_event_listener` with one that simply calls `fn`
    # right away.
    def my_add_ev_listener(event_listener: logcat_thread.EventListener):
      # Check that the event matches what's expected.
      event = event_listener.regexp
      match = event.match('score: 200.0')
      if match is None:  # Ignore events that are not scores.
        return

      event_listener.handler_fn(event, match)

      # Scores are accumulated by their differences, so a subsequent lower score
      # means that the final reward decreases.
      match = event.match('score: 185')
      event_listener.handler_fn(event, match)

    task = task_pb2.Task()
    task.log_parsing_config.log_regexps.score = (
        '^score: ([-+]?[0-9]*\\.?[0-9]*)$')
    task_mgr = task_manager.TaskManager(task=task)
    self._logcat_thread.add_event_listener.side_effect = my_add_ev_listener
    task_mgr.setup_task(
        adb_controller=self._adb_controller,
        log_stream=self._log_stream)
    self.assertEqual(task_mgr.get_current_reward(), 185.0)

  def test_get_current_extras(self):
    # Replace `LogcatThread.add_event_listener` with one that simply calls `fn`
    # right away.
    def my_add_ev_listener(event_listener: logcat_thread.EventListener):
      # Check that the event matches what's expected.
      event = event_listener.regexp
      match = event.match('extra: some_extra [1, 2]')
      if match is None:  # Ignore events that are not extras.
        return

      # Emit events.
      fn = event_listener.handler_fn
      fn(event, event.match('extra: an_extra [1, 2, 3]'))
      fn(event, event.match('extra: an_extra [4, 5, 6]'))
      fn(event, event.match('extra: another_extra 0.5'))
      fn(event, event.match('extra: multi_dimension_extra [[9,8,7],[6,5,4]]'))
      fn(event, event.match('extra: boolean_extra'))

    # Setup the task and trigger the listener.
    task = task_pb2.Task()
    task.log_parsing_config.log_regexps.extra.extend([
        '^extra: (?P<name>[^ ]*)[ ]?(?P<extra>.*)$'
    ])
    task_mgr = task_manager.TaskManager(task=task)
    self._logcat_thread.add_event_listener.side_effect = my_add_ev_listener
    task_mgr.setup_task(
        adb_controller=self._adb_controller,
        log_stream=self._log_stream)

    # Check expectations.
    extras = task_mgr.get_current_extras()
    np.testing.assert_almost_equal([[1, 2, 3], [4, 5, 6]],
                                   extras.get('an_extra'))
    np.testing.assert_almost_equal([0.5], extras.get('another_extra'))
    np.testing.assert_almost_equal([[[9, 8, 7], [6, 5, 4]]],
                                   extras.get('multi_dimension_extra'))
    np.testing.assert_equal([1], extras.get('boolean_extra'))

  def test_get_current_extras_json_format(self):
    # Replace `LogcatThread.add_event_listener` with one that simply calls `fn`
    # right away.
    def my_add_ev_listener(event_listener: logcat_thread.EventListener):
      # Check that the event matches what's expected.
      event = event_listener.regexp
      match = event.match('json_extra: {}')
      if match is None:  # Ignore events that are not extras.
        return

      # Emit events.
      extra = {
          'extra_scalar': 0,
          'extra_list': [1, 2, 3, 4],
          'extra_dict': {
              'foo': 'bar'
          },
          'extra_string': 'a_string'
      }
      extra_update = {'extra_string': 'a_new_string', 'extra_float': 0.6}
      fn = event_listener.handler_fn
      fn(event, event.match(f'json_extra: {json.dumps(extra)}'))
      fn(event, event.match(f'json_extra: {json.dumps(extra_update)}'))

    # Setup the task and trigger the listener.
    task = task_pb2.Task()
    task.log_parsing_config.log_regexps.json_extra.extend([
        '^json_extra: (?P<json_extra>.*)$'
    ])
    task_mgr = task_manager.TaskManager(task=task)
    self._logcat_thread.add_event_listener.side_effect = my_add_ev_listener
    task_mgr.setup_task(
        adb_controller=self._adb_controller,
        log_stream=self._log_stream)

    # Check expectations.
    expected_extra = {
        'extra_scalar': [0],
        'extra_list': [[1, 2, 3, 4]],
        'extra_dict': [{
            'foo': 'bar'
        }],
        'extra_string': ['a_string', 'a_new_string'],
        'extra_float': [0.6]
    }
    extras = task_mgr.get_current_extras()
    np.testing.assert_almost_equal(
        expected_extra.get('extra_scalar'), extras.get('extra_scalar'))
    np.testing.assert_almost_equal(
        expected_extra.get('extra_list'), extras.get('extra_list'))
    np.testing.assert_equal(
        expected_extra.get('extra_string'), extras.get('extra_string'))
    np.testing.assert_almost_equal(
        expected_extra.get('extra_float'), extras.get('extra_float'))
    np.testing.assert_equal(
        expected_extra.get('extra_dict'), extras.get('extra_dict'))

  def test_multi_log_regexp(self):
    # Replace `LogcatThread.add_event_listener` with one that simply calls `fn`
    # right away.
    def my_add_ev_listener(event_listener: logcat_thread.EventListener):
      # Check that the event matches what's expected.
      match = event_listener.regexp.match('Reward_2: 123.0')
      if match is None:  # Ignore events that are not rewards.
        return

      event_listener.handler_fn(event_listener.regexp, match)

    task = task_pb2.Task()
    task.log_parsing_config.log_regexps.reward.extend([
        '^[Rr]eward_1: ([-+]?[0-9]*\\.?[0-9]*)$',
        '^[Rr]eward_2: ([-+]?[0-9]*\\.?[0-9]*)$'
    ])
    task_mgr = task_manager.TaskManager(task=task)
    self._logcat_thread.add_event_listener.side_effect = my_add_ev_listener
    task_mgr.setup_task(
        adb_controller=self._adb_controller,
        log_stream=self._log_stream)
    self.assertEqual(task_mgr.get_current_reward(), 123.0)

  def test_multi_reward_regexp(self):
    # Replace `LogcatThread.add_event_listener` with one that simply calls `fn`
    # right away.'

    def my_add_ev_listener(event_listener: logcat_thread.EventListener):
      # Check that the event matches what's expected.
      match_1 = event_listener.regexp.match('Reward_1: 5.0')
      match_2 = event_listener.regexp.match('Reward_2: 10.0')

      if match_1:
        event_listener.handler_fn(event_listener.regexp, match_1)

      if match_2:
        event_listener.handler_fn(event_listener.regexp, match_2)

    task = task_pb2.Task()
    task.log_parsing_config.log_regexps.reward.extend([
        '^[Rr]eward_1: ([-+]?[0-9]*\\.?[0-9]*)$',
        '^[Rr]eward_2: ([-+]?[0-9]*\\.?[0-9]*)$',
    ])
    task_mgr = task_manager.TaskManager(task=task)
    self._logcat_thread.add_event_listener.side_effect = my_add_ev_listener
    task_mgr.setup_task(
        adb_controller=self._adb_controller,
        log_stream=self._log_stream)
    self.assertEqual(task_mgr.get_current_reward(), 15.0)

  def test_check_episode_end(self):
    # Replace `LogcatThread.add_event_listener` with one that simply calls `fn`
    # right away.
    def my_add_ev_listener(event_listener: logcat_thread.EventListener):
      # Check that the event matches what's expected.
      event = event_listener.regexp
      match = event.match('I am done!')
      if match is None:  # Ignore events that are not episode end.
        return

      event_listener.handler_fn(event, match)

    task = task_pb2.Task()
    task.log_parsing_config.log_regexps.episode_end.extend(['I am done!'])
    task_mgr = task_manager.TaskManager(task=task)
    self._logcat_thread.add_event_listener.side_effect = my_add_ev_listener
    task_mgr.setup_task(
        adb_controller=self._adb_controller,
        log_stream=self._log_stream)
    episode_end = task_mgr.check_if_episode_ended()
    self.assertTrue(episode_end)

if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for android_env.components.emulator_launcher."""

import builtins
import os
import subprocess

from absl.testing import absltest
from android_env.components.simulators.emulator import emulator_launcher
import grpc
import mock


class EmulatorLauncherTest(absltest.TestCase):

  def setUp(self):
    super().setUp()

    self._emulator_path = 'fake/path/emulator'
    self._adb_port = 5554
    self._adb_server_port = 1234
    self._emulator_console_port = 5555
    self._avd_name = 'my_avd_name'

    self._emulator_output = mock.create_autospec(open)
    self._emulator_output.close = lambda: None

    self._grpc_channel = mock.create_autospec(grpc.Channel)
    mock.patch.object(
        grpc.aio, 'secure_channel',
        return_value=self._grpc_channel).start()
    mock.patch.object(
        grpc, 'secure_channel',
        return_value=self._grpc_channel).start()
    mock.patch.object(
        grpc, 'local_channel_credentials',
        return_value=self._grpc_channel).start()
    mock.patch.object(grpc, 'channel_ready_future').start()

    self._expected_command = [
        self._emulator_path,
        '-no-snapshot',
        '-gpu', 'swiftshader_indirect',
        '-no-audio',
        '-verbose',
        '-avd', self._avd_name,
    ]
    self._ports = ['-ports', f'{self._emulator_console_port},{self._adb_port}']

    base_lib_dir = self._emulator_path[:-8] + 'lib64/'
    ld_library_path = ':'.join([
        base_lib_dir + 'x11/',
        base_lib_dir + 'qt/lib/',
        base_lib_dir + 'gles_swiftshader/',
        base_lib_dir
    ])

    self._expected_env_vars = {
        'ANDROID_HOME': '',
        'ANDROID_SDK_ROOT': '',
        'ANDROID_AVD_HOME': '',
        'ANDROID_EMULATOR_KVM_DEVICE': '/dev/kvm',
        'ANDROID_ADB_SERVER_PORT': '1234',
        'LD_LIBRARY_PATH': ld_library_path,
        'QT_DEBUG_PLUGINS': '1',
        'QT_XKB_CONFIG_ROOT': str(self._emulator_path[:-8] + 'qt_config/'),
    }

  @mock.patch.object(os, 'environ', autospec=True, return_value=dict())
  def test_launch(self, os_environ):
    del os_environ

    launcher = emulator_launcher.EmulatorLauncher(
        adb_port=self._adb_port,
        adb_server_port=self._adb_server_port,
        emulator_console_port=self._emulator_console_port,
        emulator_path=self._emulator_path,
        avd_name=self._avd_name,
        grpc_port=-1)

    with mock.patch.object(
        subprocess, 'Popen', autospec=True) as emulator_init, \
        mock.patch.object(
            builtins, 'open', autospec=True,
            return_value=self._emulator_output):

      launcher.launch()
      emulator_init.assert_called_once_with(
          args=self._expected_command + self._ports,
          env=self._expected_env_vars,
          stdout=self._emulator_output,
          stderr=self._emulator_output)

  @mock.patch.object(os, 'environ', autospec=True, return_value=dict())
  def test_grpc_port(self, os_environ):
    del os_environ

    launcher = emulator_launcher.EmulatorLauncher(
        adb_port=self._adb_port,
        adb_server_port=self._adb_server_port,
        emulator_console_port=self._emulator_console_port,
        emulator_path=self._emulator_path,
        avd_name=self._avd_name,
        grpc_port=8554)

    with mock.patch.object(
        subprocess, 'Popen', autospec=True) as emulator_init, \
        mock.patch.object(
            builtins, 'open', autospec=True,
            return_value=self._emulator_output):
      launcher.launch()
      emulator_init.assert_called_once_with(
          args=self._expected_command + ['-grpc', '8554'] + self._ports,
          env=self._expected_env_vars,
          stdout=self._emulator_output,
          stderr=self._emulator_output)

  @mock.patch.object(os, 'environ', autospec=True, return_value=dict())
  def test_restart(self, os_environ):
    del os_environ

    launcher = emulator_launcher.EmulatorLauncher(
        adb_port=self._adb_port,
        adb_server_port=self._adb_server_port,
        emulator_console_port=self._emulator_console_port,
        emulator_path=self._emulator_path,
        avd_name=self._avd_name,
        grpc_port=-1)

    with mock.patch.object(
        subprocess, 'Popen', autospec=True) as emulator_init, \
        mock.patch.object(
            builtins, 'open', autospec=True,
            return_value=self._emulator_output):
      launcher.launch()
      launcher.restart()
      emulator_init.assert_has_calls([
          # 1st call is the first boot process.
          mock.call(
              args=self._expected_command + self._ports,
              env=self._expected_env_vars,
              stdout=self._emulator_output,
              stderr=self._emulator_output),
          # 2nd call is for waiting for the process to finish gracefully.
          mock.call().wait(timeout=30.0),
          # 3rd call is for the actual restart.
          mock.call(
              args=self._expected_command + self._ports,
              env=self._expected_env_vars,
              stdout=self._emulator_output,
              stderr=self._emulator_output)
      ])


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for log_stream."""

from absl.testing import absltest
from android_env.components import log_stream


class FakeLogStream(log_stream.LogStream):

  def __init__(self, filter_name: str):
    super().__init__()
    self._filter_name = filter_name

  def _get_stream_output(self):
    """Starts a log process and returns the stream of logs."""
    lines = [
        f'{self._filter_name} fake_line_1',
        'fake_line_2',
        f'{self._filter_name} fake_line_3',
        f'{self._filter_name} fake_line_4',
        'fake_line_5',
        'fake_line_6',
    ]
    for line in lines:
      if f'{self._filter_name}:V' in self._filters:
        if self._filter_name in line:
          yield line
      else:
        yield line

  def stop_stream(self):
    """Stops the log stream from the simulator."""
    pass


class LogStreamTest(absltest.TestCase):

  def test_get_stream_output(self):
    filter_name = 'AndroidRLTask'
    stream = FakeLogStream(filter_name=filter_name)
    stream_output = stream.get_stream_output()
    expected_lines = [
        f'{filter_name} fake_line_1',
        'fake_line_2',
        f'{filter_name} fake_line_3',
        f'{filter_name} fake_line_4',
        'fake_line_5',
        'fake_line_6',
    ]
    for line, expected_line in zip(stream_output, expected_lines):
      self.assertEqual(line, expected_line)

  def test_set_log_filters(self):
    filter_name = 'AndroidRLTask'
    stream = FakeLogStream(filter_name=filter_name)
    stream.set_log_filters([f'{filter_name}:V'])
    stream_output = stream.get_stream_output()
    expected_lines = [
        f'{filter_name} fake_line_1',
        f'{filter_name} fake_line_3',
        f'{filter_name} fake_line_4',
    ]
    for line, expected_line in zip(stream_output, expected_lines):
      self.assertEqual(line, expected_line)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for android_env.components.setup_step_interpreter."""

from absl.testing import absltest
from android_env.components import adb_controller
from android_env.components import errors
from android_env.components import logcat_thread
from android_env.components import setup_step_interpreter
from android_env.proto import adb_pb2
from android_env.proto import task_pb2
import mock

from google.protobuf import text_format


def _to_proto(proto_class, text):
  proto = proto_class()
  text_format.Parse(text, proto)
  return proto


class SetupStepInterpreterTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.logcat = mock.create_autospec(logcat_thread.LogcatThread)
    self.adb_controller = mock.create_autospec(adb_controller.AdbController)

  def test_empty_setup_steps(self):
    """Simple test where nothing should break, and nothing should be done.

    The test simply expects this test to not crash.
    """
    interpreter = setup_step_interpreter.SetupStepInterpreter(
        adb_controller=self.adb_controller, logcat=self.logcat)
    interpreter.interpret([])

  def test_none_setup_steps(self):
    """Simple test where nothing should break, and nothing should be done.

    The test simply expects this test to not crash.
    """
    interpreter = setup_step_interpreter.SetupStepInterpreter(
        adb_controller=self.adb_controller, logcat=self.logcat)
    # Empty setup steps should be ignored.
    interpreter.interpret([None])

  def test_invalid_setup_step(self):
    interpreter = setup_step_interpreter.SetupStepInterpreter(
        adb_controller=self.adb_controller, logcat=self.logcat)
    # Empty setup steps should be ignored.
    with self.assertRaises(AssertionError):
      interpreter.interpret([_to_proto(task_pb2.SetupStep, '')])

  def test_adb_install_apk_filesystem(self):

    interpreter = setup_step_interpreter.SetupStepInterpreter(
        adb_controller=self.adb_controller, logcat=self.logcat)
    interpreter.interpret([
        _to_proto(
            task_pb2.SetupStep, """
adb_call: {
  install_apk: {
    filesystem: {
      path: "/my/favorite/dir/my_apk.apk"
    }
  }
}""")
    ])
    self.adb_controller.install_apk.assert_called_once_with(
        '/my/favorite/dir/my_apk.apk')

  def test_adb_force_stop(self):
    interpreter = setup_step_interpreter.SetupStepInterpreter(
        adb_controller=self.adb_controller, logcat=self.logcat)
    interpreter.interpret([
        _to_proto(
            task_pb2.SetupStep, """
adb_call: { force_stop: { package_name: "my.app.Activity" } }""")
    ])
    # AdbController should be called exactly once with the following arguments.
    self.adb_controller.force_stop.assert_called_once_with('my.app.Activity')

  def test_adb_clear_cache(self):
    interpreter = setup_step_interpreter.SetupStepInterpreter(
        adb_controller=self.adb_controller, logcat=self.logcat)
    interpreter.interpret([
        _to_proto(
            task_pb2.SetupStep, """
adb_call: { clear_cache: { package_name: "my.app.Activity" } }""")
    ])
    # AdbController should be called exactly once with the following arguments.
    self.adb_controller.clear_cache.assert_called_once_with('my.app.Activity')

  def test_adb_grant_permissions(self):

    interpreter = setup_step_interpreter.SetupStepInterpreter(
        adb_controller=self.adb_controller, logcat=self.logcat)
    interpreter.interpret([
        _to_proto(
            task_pb2.SetupStep, """
adb_call: {
  grant_permissions: {
    package_name: "my.app.Activity"
    permissions: [ "my.namespace.READ_DATA", "another.namespace.WRITE" ]
  }
}""")
    ])
    # AdbController should be called exactly once with the following arguments.
    self.adb_controller.grant_permissions.assert_called_once_with(
        'my.app.Activity',
        ['my.namespace.READ_DATA', 'another.namespace.WRITE'])

  def test_adb_start_activity(self):
    interpreter = setup_step_interpreter.SetupStepInterpreter(
        adb_controller=self.adb_controller, logcat=self.logcat)
    interpreter.interpret([
        _to_proto(
            task_pb2.SetupStep, """
adb_call: {
  start_activity: {
    full_activity: "my.app.Activity"
    extra_args: "arg1"
    extra_args: "arg2"
  }
}""")
    ])
    # AdbController should be called exactly once with the following arguments.
    self.adb_controller.start_activity.assert_called_once_with(
        'my.app.Activity', ['arg1', 'arg2'])

  def test_adb_single_tap(self):
    interpreter = setup_step_interpreter.SetupStepInterpreter(
        adb_controller=self.adb_controller, logcat=self.logcat)
    interpreter.interpret([
        _to_proto(task_pb2.SetupStep, """
adb_call: {
  tap: {
    x: 321
    y: 654
  }
}""")
    ])
    # AdbController should be called exactly once with the following arguments.
    self.adb_controller.input_tap.assert_called_once_with(321, 654)

  def test_adb_rotate(self):
    interpreter = setup_step_interpreter.SetupStepInterpreter(
        adb_controller=self.adb_controller, logcat=self.logcat)
    # Check landscape.
    interpreter.interpret([
        _to_proto(task_pb2.SetupStep,
                  """ adb_call: { rotate: { orientation: LANDSCAPE_90 } }""")
    ])
    # AdbController should be called exactly once with the following arguments.
    self.adb_controller.rotate_device.assert_called_once_with(
        adb_pb2.AdbCall.Rotate.Orientation.LANDSCAPE_90)

    self.adb_controller.reset_mock()
    # Check portrait.
    interpreter.interpret([
        _to_proto(task_pb2.SetupStep,
                  """ adb_call: { rotate: { orientation: PORTRAIT_0 } }""")
    ])
    # AdbController should be called exactly once with the following arguments.
    self.adb_controller.rotate_device.assert_called_once_with(
        adb_pb2.AdbCall.Rotate.Orientation.PORTRAIT_0)

    self.adb_controller.reset_mock()
    # Check landscape inverted.
    interpreter.interpret([
        _to_proto(task_pb2.SetupStep,
                  """ adb_call: { rotate: { orientation: LANDSCAPE_270} }""")
    ])
    # AdbController should be called exactly once with the following arguments.
    self.adb_controller.rotate_device.assert_called_once_with(
        adb_pb2.AdbCall.Rotate.Orientation.LANDSCAPE_270)

    self.adb_controller.reset_mock()
    # Check portrait up-side-down.
    interpreter.interpret([
        _to_proto(task_pb2.SetupStep,
                  """ adb_call: { rotate: { orientation: PORTRAIT_180 } }""")
    ])
    # AdbController should be called exactly once with the following arguments.
    self.adb_controller.rotate_device.assert_called_once_with(
        adb_pb2.AdbCall.Rotate.Orientation.PORTRAIT_180)

  def test_adb_press_button(self):
    interpreter = setup_step_interpreter.SetupStepInterpreter(
        adb_controller=self.adb_controller, logcat=self.logcat)
    interpreter.interpret([
        _to_proto(task_pb2.SetupStep,
                  """ adb_call: { press_button: { button: HOME } }""")
    ])
    # AdbController should be called exactly once with the following arguments.
    self.adb_controller.input_key.assert_called_once_with('KEYCODE_HOME')
    self.adb_controller.reset_mock()
    interpreter.interpret([
        _to_proto(task_pb2.SetupStep,
                  """ adb_call: { press_button: { button: BACK } }""")
    ])
    # AdbController should be called exactly once with the following arguments.
    self.adb_controller.input_key.assert_called_once_with('KEYCODE_BACK')

  def test_adb_start_accessibility_service(self):
    interpreter = setup_step_interpreter.SetupStepInterpreter(
        adb_controller=self.adb_controller, logcat=self.logcat)
    interpreter.interpret([
        _to_proto(
            task_pb2.SetupStep, """
adb_call: {
  start_accessibility_service: {
    full_service: "my.app.AccessibilityService"
  }
}""")
    ])
    # AdbController should be called exactly once with the following arguments.
    self.adb_controller.start_accessibility_service.assert_called_once_with(
        'my.app.AccessibilityService')

  def test_adb_start_screen_pinning(self):
    interpreter = setup_step_interpreter.SetupStepInterpreter(
        adb_controller=self.adb_controller, logcat=self.logcat)
    interpreter.interpret([
        _to_proto(
            task_pb2.SetupStep, """
adb_call: {
  start_screen_pinning: {
    full_activity: "my.app.HighlanderApp"  # "There can be only one".
  }
}""")
    ])
    # AdbController should be called once with the following arguments.
    self.adb_controller.start_screen_pinning.assert_called_with(
        u'my.app.HighlanderApp')

  @mock.patch('time.sleep')
  def test_time_sleep(self, mock_sleep):
    interpreter = setup_step_interpreter.SetupStepInterpreter(
        adb_controller=self.adb_controller, logcat=self.logcat)
    interpreter.interpret(
        [_to_proto(task_pb2.SetupStep, """sleep: { time_sec: 0.875 }""")])
    assert mock_sleep.call_count == 2
    mock_sleep.assert_has_calls([mock.call(0.875), mock.call(0.5)])

  @mock.patch('time.sleep')
  def test_wait_for_app_screen_empty_activity(self, unused_mock_sleep):
    interpreter = setup_step_interpreter.SetupStepInterpreter(
        adb_controller=self.adb_controller, logcat=self.logcat)
    with self.assertRaises(errors.StepCommandError):
      interpreter.interpret([
          _to_proto(task_pb2.SetupStep,
                    """success_condition: {wait_for_app_screen: { }}""")
      ])

  def test_wait_for_message_fail(self):
    interpreter = setup_step_interpreter.SetupStepInterpreter(
        adb_controller=self.adb_controller, logcat=self.logcat)
    self.assertRaises(errors.StepCommandError, interpreter.interpret, [
        _to_proto(
            task_pb2.SetupStep, """
success_condition: {
  wait_for_message: {
    message:'foo'
    timeout_sec: 0.0001
  }
}
""")
    ])

  def test_wait_for_message_success(self):
    interpreter = setup_step_interpreter.SetupStepInterpreter(
        adb_controller=self.adb_controller, logcat=self.logcat)

    # Replace `LogcatThread.add_event_listener` with one that simply calls `fn`
    # right away, ignoring `event`.
    def mock_add_ev_listener(event_listener):
      event_listener.handler_fn('some_event', 'some_match')

    self.logcat.add_event_listener.side_effect = mock_add_ev_listener
    # The test checks that this command raises no AssertionError.
    interpreter.interpret([
        _to_proto(
            task_pb2.SetupStep, """
success_condition: {
  wait_for_message: {
    message:'foo'
    timeout_sec: 1.0
  }
}
""")
    ])

  @mock.patch('time.sleep')
  def test_check_install_not_installed(self, unused_mock_sleep):
    self.adb_controller.is_package_installed.return_value = False
    interpreter = setup_step_interpreter.SetupStepInterpreter(
        adb_controller=self.adb_controller, logcat=self.logcat)
    with self.assertRaises(errors.StepCommandError):
      interpreter.interpret([
          _to_proto(
              task_pb2.SetupStep, """
success_condition: {
  check_install: {
    package_name: "faz"
    timeout_sec: 0.0001
  }
}
""")
      ])

  def test_check_install_installed(self):
    self.adb_controller.is_package_installed.return_value = True
    interpreter = setup_step_interpreter.SetupStepInterpreter(
        adb_controller=self.adb_controller, logcat=self.logcat)
    # The test checks that this command raises no AssertionError.
    interpreter.interpret([
        _to_proto(
            task_pb2.SetupStep, """
success_condition: {
  check_install: {
    package_name: "baz"
    timeout_sec: 0.0001
  }
}""")
    ])
    self.adb_controller.is_package_installed.assert_called_once_with('baz')

  def test_num_retries_failure(self):
    self.adb_controller.is_package_installed.side_effect = [False] * 3
    interpreter = setup_step_interpreter.SetupStepInterpreter(
        adb_controller=self.adb_controller, logcat=self.logcat)
    with self.assertRaises(errors.StepCommandError):
      interpreter.interpret([
          _to_proto(
              task_pb2.SetupStep, """
success_condition: {
  check_install: {
    package_name: "faz"
    timeout_sec: 0.0001
  }
  num_retries: 3
}""")
      ])
    # We retried 3 times after the first call, so we expect 3+1 calls.
    self.assertEqual(3, self.adb_controller.is_package_installed.call_count)

  @mock.patch('time.sleep')
  def test_num_retries_success(self, unused_mock_sleep):
    self.adb_controller.is_package_installed.side_effect = [
        False, False, True, False
    ]
    interpreter = setup_step_interpreter.SetupStepInterpreter(
        adb_controller=self.adb_controller, logcat=self.logcat)
    interpreter.interpret([
        _to_proto(
            task_pb2.SetupStep, """
success_condition: {
  check_install: {
    package_name: "bar"
    timeout_sec: 0.0001
  }
  num_retries: 5
}""")
    ])
    # The check should succeed on the third try.
    self.assertEqual(3, self.adb_controller.is_package_installed.call_count)

  def test_retry_step(self):
    self.adb_controller.is_package_installed.side_effect = [False, True]
    interpreter = setup_step_interpreter.SetupStepInterpreter(
        adb_controller=self.adb_controller, logcat=self.logcat)
    interpreter.interpret([
        _to_proto(
            task_pb2.SetupStep, """
adb_call: { press_button: { button: HOME } }
success_condition: {
  check_install: {
    package_name: "bar"
    timeout_sec: 0.0001
  }
  num_retries: 2
}""")
    ])
    # We expect the check to fail twice and succeed on the third pass.
    self.adb_controller.input_key.assert_has_calls(
        [mock.call('KEYCODE_HOME')] * 2)
    self.assertEqual(2, self.adb_controller.is_package_installed.call_count)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for specs.py."""

from absl.testing import absltest
from absl.testing import parameterized
from android_env.components import specs
from android_env.proto import task_pb2
from dm_env import specs as dm_env_specs
import numpy as np


class SpecsTest(parameterized.TestCase):

  def test_base_action_spec(self):
    action_spec = specs.base_action_spec(num_fingers=1)
    for spec in action_spec.values():
      self.assertIsInstance(spec, dm_env_specs.Array)
    self.assertEqual(action_spec['action_type'].shape, ())
    self.assertEqual(action_spec['action_type'].dtype, np.int32)
    self.assertEqual(action_spec['touch_position'].shape, (2,))
    self.assertEqual(action_spec['touch_position'].dtype, np.float32)

  def test_base_action_spec_multitouch(self):
    action_spec = specs.base_action_spec(num_fingers=3)
    self.assertLen(action_spec.keys(), 6)
    for spec in action_spec.values():
      self.assertIsInstance(spec, dm_env_specs.Array)
    self.assertEqual(action_spec['action_type'].shape, ())
    self.assertEqual(action_spec['action_type'].dtype, np.int32)
    self.assertEqual(action_spec['touch_position'].shape, (2,))
    self.assertEqual(action_spec['touch_position'].dtype, np.float32)
    self.assertEqual(action_spec['action_type_2'].shape, ())
    self.assertEqual(action_spec['action_type_2'].dtype, np.int32)
    self.assertEqual(action_spec['touch_position_2'].shape, (2,))
    self.assertEqual(action_spec['touch_position_2'].dtype, np.float32)
    self.assertEqual(action_spec['action_type_3'].shape, ())
    self.assertEqual(action_spec['action_type_3'].dtype, np.int32)
    self.assertEqual(action_spec['touch_position_3'].shape, (2,))
    self.assertEqual(action_spec['touch_position_3'].dtype, np.float32)

  @parameterized.parameters(
      (480, 320),
      (100, 100),
      (1440, 1960),
  )
  def test_base_observation_spec(self, height, width):
    observation_spec = specs.base_observation_spec(height, width)
    for spec in observation_spec.values():
      self.assertIsInstance(spec, dm_env_specs.Array)
    self.assertEqual(observation_spec['pixels'].shape, (height, width, 3))
    self.assertEqual(observation_spec['pixels'].dtype, np.uint8)
    self.assertEqual(observation_spec['timedelta'].shape, ())
    self.assertEqual(observation_spec['timedelta'].dtype, np.int64)
    self.assertEqual(observation_spec['orientation'].shape, (4,))
    self.assertEqual(observation_spec['orientation'].dtype, np.uint8)

  def test_base_task_extras_spec(self):
    array_spec_1 = task_pb2.ArraySpec()
    array_spec_1.name = 'my_extra_1'
    array_spec_1.shape.extend([10, 10])
    array_spec_1.dtype = task_pb2.ArraySpec.FLOAT

    array_spec_2 = task_pb2.ArraySpec()
    array_spec_2.name = 'my_extra_2'
    array_spec_2.shape.extend([1])
    array_spec_2.dtype = task_pb2.ArraySpec.UINT8

    fake_task = task_pb2.Task()
    fake_task.extras_spec.extend([array_spec_1, array_spec_2])
    task_extras_spec = specs.base_task_extras_spec(fake_task)
    for spec in task_extras_spec.values():
      self.assertIsInstance(spec, dm_env_specs.Array)

    self.assertEqual(task_extras_spec['my_extra_1'].shape, (10, 10))
    self.assertEqual(task_extras_spec['my_extra_2'].shape, (1,))
    self.assertEqual(task_extras_spec['my_extra_1'].dtype, np.float32)
    self.assertEqual(task_extras_spec['my_extra_2'].dtype, np.uint8)

  @parameterized.parameters(
      ('name_1', [480, 320, 3], task_pb2.ArraySpec.FLOAT, np.float32),
      ('name_2', [100, 100, 3], task_pb2.ArraySpec.INT32, np.int32),
      ('name_3', [123, 456, 3], task_pb2.ArraySpec.UINT8, np.uint8),
      ('name_4', [480, 320, 1], task_pb2.ArraySpec.BOOL, np.bool_),
      ('', [480, 320], task_pb2.ArraySpec.STRING_U25, np.dtype(('<U25'))),
  )
  def test_convert_spec(self, name, shape, dtype, expected_dtype):
    fake_array_spec = task_pb2.ArraySpec()
    fake_array_spec.name = name
    fake_array_spec.dtype = dtype
    fake_array_spec.shape.extend(shape)
    fake_task = task_pb2.Task()
    fake_task.extras_spec.extend([fake_array_spec])
    task_extras_spec = specs.base_task_extras_spec(fake_task)
    for spec in task_extras_spec.values():
      self.assertIsInstance(spec, dm_env_specs.Array)

    self.assertEqual(task_extras_spec[name].shape, tuple(shape))
    self.assertEqual(task_extras_spec[name].dtype, expected_dtype)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for android_env.components.emulator_simulator."""

from absl.testing import absltest
from android_env.components import action_type
from android_env.components import adb_controller
from android_env.components.simulators.emulator import emulator_launcher
from android_env.components.simulators.emulator import emulator_simulator
import grpc
import mock
import numpy as np
from PIL import Image

from android_env.proto import emulator_controller_pb2
from android_env.proto import emulator_controller_pb2_grpc


class EmulatorSimulatorTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.addCleanup(mock.patch.stopall)  # Disable previous patches.

    self._adb_controller = mock.create_autospec(adb_controller.AdbController)
    self._launcher = mock.create_autospec(emulator_launcher.EmulatorLauncher)
    self._emulator_stub = mock.create_autospec(
        emulator_controller_pb2_grpc.EmulatorControllerStub)
    self._grpc_channel = mock.create_autospec(grpc.Channel)

    self._grpc_channel = mock.create_autospec(grpc.Channel)
    mock.patch.object(
        grpc.aio, 'insecure_channel',
        return_value=self._grpc_channel).start()
    mock.patch.object(
        grpc, 'insecure_channel',
        return_value=self._grpc_channel).start()
    mock.patch.object(
        adb_controller, 'AdbController',
        return_value=self._adb_controller).start()
    mock.patch.object(
        emulator_launcher, 'EmulatorLauncher',
        return_value=self._launcher).start()

  def test_adb_device_name_not_empty(self):
    tmp_dir = absltest.get_default_test_tmpdir()
    simulator = emulator_simulator.EmulatorSimulator(
        num_fingers=1,
        tmp_dir=tmp_dir,
        emulator_launcher_args={'grpc_port': 1234},
        adb_controller_args={
            'adb_path': '/my/adb',
            'adb_server_port': 5037,
            'prompt_regex': 'awesome>',
        })
    self.assertNotEmpty(simulator.adb_device_name())

  def test_launch(self):

    # Make sure that adb_controller is started before Emulator is launched.
    call_order = []
    self._adb_controller.init_server.side_effect = (
        lambda *a, **kw: call_order.append('init_server'))
    self._launcher.launch.side_effect = (
        lambda *a, **kw: call_order.append('launch'))

    tmp_dir = absltest.get_default_test_tmpdir()
    simulator = emulator_simulator.EmulatorSimulator(
        num_fingers=1,
        tmp_dir=tmp_dir,
        emulator_launcher_args={'grpc_port': 1234},
        adb_controller_args={
            'adb_path': '/my/adb',
            'adb_server_port': 5037,
            'prompt_regex': 'awesome>',
        })
    self._adb_controller.get_screen_dimensions.return_value = (1234, 5678)

    # The simulator should launch and not crash.
    simulator.launch()
    self.assertEqual(call_order, ['init_server', 'launch'])

  def test_close(self):
    tmp_dir = absltest.get_default_test_tmpdir()
    simulator = emulator_simulator.EmulatorSimulator(
        num_fingers=1,
        tmp_dir=tmp_dir,
        emulator_launcher_args={'grpc_port': 1234},
        adb_controller_args={
            'adb_path': '/my/adb',
            'adb_server_port': 5037,
            'prompt_regex': 'awesome>',
        })

    # The simulator should launch and not crash.
    self._adb_controller.get_screen_dimensions.return_value = (1234, 5678)
    simulator.launch()

    # For whatever reason clients may want to close the EmulatorSimulator.
    # We just want to check that the simulator does not crash and/or leak
    # resources.
    simulator.close()

  def test_restart(self):
    tmp_dir = absltest.get_default_test_tmpdir()
    simulator = emulator_simulator.EmulatorSimulator(
        num_fingers=1,
        tmp_dir=tmp_dir,
        emulator_launcher_args={'grpc_port': 1234},
        adb_controller_args={
            'adb_path': '/my/adb',
            'adb_server_port': 5037,
            'prompt_regex': 'awesome>',
        })

    # The simulator should launch and not crash.
    self._adb_controller.get_screen_dimensions.return_value = (1234, 5678)
    simulator.launch()

    # For whatever reason clients may want to restart the EmulatorSimulator.
    simulator.restart()

  def test_get_observation(self):
    tmp_dir = absltest.get_default_test_tmpdir()
    simulator = emulator_simulator.EmulatorSimulator(
        num_fingers=1,
        tmp_dir=tmp_dir,
        emulator_launcher_args={'grpc_port': 1234},
        adb_controller_args={
            'adb_path': '/my/adb',
            'adb_server_port': 5037,
            'prompt_regex': 'awesome>',
        })

    # The simulator should launch and not crash.
    self._adb_controller.get_screen_dimensions.return_value = (1234, 5678)
    simulator.launch()

    simulator._emulator_stub.getScreenshot = mock.MagicMock(
        return_value=emulator_controller_pb2.Image(
            format=emulator_controller_pb2.ImageFormat(width=5678, height=1234),
            image=Image.new('RGB', (1234, 5678)).tobytes(),
            timestampUs=123))

    observation = simulator.get_observation()
    # The observation should have three components:
    #   - an image
    #   - the timedelta
    #   - the orientation.
    self.assertLen(observation, 3)
    # The first element (the "image") should have the same screen dimensions as
    # reported by ADB and it should have 3 channels (RGB).
    self.assertEqual(observation['pixels'].shape, (1234, 5678, 3))
    self.assertEqual(observation['timedelta'], 123)

  def test_send_action(self):
    tmp_dir = absltest.get_default_test_tmpdir()
    simulator = emulator_simulator.EmulatorSimulator(
        num_fingers=1,
        tmp_dir=tmp_dir,
        emulator_launcher_args={'grpc_port': 1234},
        adb_controller_args={
            'adb_path': '/my/adb',
            'adb_server_port': 5037,
            'prompt_regex': 'awesome>',
        })

    # The simulator should launch and not crash.
    self._adb_controller.get_screen_dimensions.return_value = (5000, 1000)
    simulator.launch()

    simulator._emulator_stub.sendTouch = mock.MagicMock(return_value=None)

    simulator.send_action(
        {'action_type': np.array([action_type.ActionType.TOUCH]),
         'touch_position': np.array([0.25, 0.75])})
    simulator.send_action(
        {'action_type': np.array([action_type.ActionType.TOUCH]),
         'touch_position': np.array([0.75, 0.50])})
    simulator.send_action(
        {'action_type': np.array([action_type.ActionType.LIFT]),
         'touch_position': np.array([0.66, 0.33])})
    # We expect EmulatorSimulator to send the following calls:
    # 1st call:
    #     x-coordinate: 10000 * 0.25 = 250
    #     y-coordinate: 50000 * 0.75 = 3750
    #     down: True  # It's a touch command.
    # 2nd call:
    #     x-coordinate: 10000 * 0.75 = 750
    #     y-coordinate: 50000 * 0.50 = 2500
    #     down: True  # It's a touch command.
    # 3rd call:
    #     x-coordinate: 0
    #     y-coordinate: 0
    #     down: False  # It's a lift command.

  def test_multitouch_action(self):
    tmp_dir = absltest.get_default_test_tmpdir()
    simulator = emulator_simulator.EmulatorSimulator(
        num_fingers=3,
        tmp_dir=tmp_dir,
        emulator_launcher_args={'grpc_port': 1234},
        adb_controller_args={
            'adb_path': '/my/adb',
            'adb_server_port': 5037,
            'prompt_regex': 'awesome>',
        })
    # The simulator should launch and not crash.
    self._adb_controller.get_screen_dimensions.return_value = (1000, 1000)
    simulator.launch()

    simulator._emulator_stub.sendTouch.return_value = None

    simulator.send_action({
        'action_type': np.array([action_type.ActionType.TOUCH]),
        'touch_position': np.array([0.25, 0.75]),
        'action_type_2': np.array([action_type.ActionType.TOUCH]),
        'touch_position_2': np.array([0.75, 0.25]),
        'action_type_3': np.array([action_type.ActionType.LIFT]),
        'touch_position_3': np.array([0.5, 0.5]),
    })

    expected_touch_event = emulator_controller_pb2.TouchEvent(touches=[
        emulator_controller_pb2.Touch(x=250, y=750, pressure=1, identifier=0),
        emulator_controller_pb2.Touch(x=750, y=250, pressure=1, identifier=1),
        emulator_controller_pb2.Touch(x=500, y=500, pressure=0, identifier=2),
    ])

    simulator._emulator_stub.sendTouch.assert_called_with(expected_touch_event)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 SJTU X-Lance Lab
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Created by Danyang Zhang @X-Lance.
"""

import random
import torch

#from absl import logging

def text_detector(screen, bboxes):
    #  function `text_detector` {{{ # 
    """
    screen - tensor of float32 with shape (3, height, width)
    bboxes - list of tensor of float32 with shape (1, 4)

    return list of list of str
    """

    decides = random.random()
    event_index = random.randrange(len(bboxes)) if decides<=0.2 else None
    #if decides<=0.2:
        #logging.info("\x1b[32;1;43mText Detected!\x1b[0m")
    return [["Event String"] if i==event_index else ["Test String"] for i, _ in enumerate(bboxes)]
    #  }}} function `text_detector` # 

def text_recognizer(screen, bboxes):
    #  function `text_recognizer` {{{ # 
    """
    screen - tensor of float32 with shape (3, height, width)
    bboxes - list of tensor of float32 with shape (1, 4)

    return list of str
    """

    decides = random.random()
    event_index = random.randrange(len(bboxes)) if decides<=0.2 else None
    #if decides<=0.2:
        #logging.info("Text Reward!")
    return ["Event String" if i==event_index else "Test String" for i, _ in enumerate(bboxes)]
    #  }}} function `text_recognizer` # 

def icon_detector(screen, bboxes):
    #  function `icon_detector` {{{ # 
    """
    screen - tensor of float32 with shape (3, height, width)
    bboxes - list of tensor of float32 with shape (1, 4) with length nb_bboxes

    return
    - tensor of float32 with shape (nb_bboxes, nb_candidates, 4)
    - list of [[list of str with length nb_candidates]] with length nb_bboxes
    """

    decides = random.random()
    event_index = random.randrange(len(bboxes)) if decides<=0.5 else None
    #if decides<=0.5:
        #logging.info("Icon Reward!")
    return torch.stack(bboxes),\
            [["Event String"] if i==event_index else ["Test String"] for i, _ in enumerate(bboxes)]
    #  }}} function `icon_detector` # 

def icon_recognizer(screen, bboxes):
    #  function `icon_recognizer` {{{ # 
    """
    screen - tensor of float32 with shape (3, height, width)
    bboxes - list of tensor of float32 with shape (1, 4)

    return list of str
    """

    decides = random.random()
    event_index = random.randrange(len(bboxes)) if decides<=0.1 else None
    return ["Event String" if i==event_index else "Test String" for i, _ in enumerate(bboxes)]
    #  }}} function `icon_recognizer` # 

def icon_matcher(screen, targets, bboxes):
    #  function `icon_matcher` {{{ # 
    """
    screen - tensor of float32 with shape (3, height, width)
    targets - list of tensor of float32 with shape (3, height', width') with
      length nb_bboxes
    bboxes - tensor of float32 with shape (nb_bboxes, nb_candidates, 4)

    return list of [[list of bool with length nb_candidates]] with length
      nb_bboxes
    """

    decides = random.random()
    event_index = random.randrange(len(bboxes)) if decides<=0.1 else None
    return [[i==event_index] + [False for _ in bb] for i, bb in enumerate(bboxes)]
    #  }}} function `icon_matcher` # 



# coding=utf-8
# vim: set tabstop=2 shiftwidth=2:
# Copyright 2023 SJTU X-Lance Lab
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Revised by Danyang Zhang @X-Lance based on
# 
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A class that manages an Android Emulator."""

import functools
import tempfile
from typing import Any, Dict, List, Optional

from absl import logging
from android_env.components import adb_controller
from android_env.components import adb_log_stream
from android_env.components import errors
from android_env.components import log_stream
from android_env.components.simulators import base_simulator
from android_env.components.simulators.emulator import emulator_launcher
import numpy as np
import portpicker

from android_env.proto import emulator_controller_pb2


def is_existing_emulator_provided(launcher_args: Dict[str, Any]) -> bool:
  """Returns true if all necessary args were provided."""
  return bool(launcher_args.get('adb_port') and
              launcher_args.get('emulator_console_port') and
              launcher_args.get('grpc_port'))


class EmulatorSimulator(base_simulator.BaseSimulator):
  """Controls an Android Emulator."""

  def __init__(self,
               emulator_launcher_args: Dict[str, Any],
               adb_controller_args: Dict[str, Any],
               tmp_dir: Optional[str] = None,
               adb_root: bool = False,
               frida_server: Optional[str] = None,
               **kwargs):

    # If adb_port, console_port and grpc_port are all already provided,
    # we assume the emulator already exists and there's no need to launch.
    if is_existing_emulator_provided(emulator_launcher_args):
      self._existing_emulator_provided = True
      self._adb_port = emulator_launcher_args['adb_port']
      self._console_port = emulator_launcher_args['emulator_console_port']
      self._grpc_port = emulator_launcher_args['grpc_port']
      logging.info('Connecting to existing emulator "%r"', self._adb_port)
    else:
      self._existing_emulator_provided = False
      self._adb_port = portpicker.pick_unused_port()
      self._console_port = portpicker.pick_unused_port()
      self._grpc_port = portpicker.pick_unused_port()

    self._emulator_stub = None
    self._image_format = None

    self._adb_root: bool = adb_root
    self._frida_server: Optional[str] = frida_server

    super().__init__(**kwargs)

    # Create directory for tmp files.
    self._tmp_dir = tmp_dir or tempfile.gettempdir()
    self._local_tmp_dir_handle = tempfile.TemporaryDirectory(dir=self._tmp_dir)
    self._local_tmp_dir = self._local_tmp_dir_handle.name
    logging.info('Simulator local_tmp_dir: %s', self._local_tmp_dir)

    # Initialize own ADB controller
    self._adb_controller_args = adb_controller_args
    self._adb_controller = self.create_adb_controller()
    self._adb_controller.init_server()
    logging.info('Initialized simulator with ADB server port %r.',
                 self._adb_controller_args['adb_server_port'])

    # Create EmulatorLauncher.
    emulator_launcher_args.update({
        'adb_port': self._adb_port,
        'adb_server_port': self._adb_controller_args['adb_server_port'],
        'emulator_console_port': self._console_port,
        'local_tmp_dir': self._local_tmp_dir,
        'grpc_port': self._grpc_port,
    })
    logging.info('emulator_launcher_args: %r', emulator_launcher_args)
    if not self._existing_emulator_provided:
      self._launcher = emulator_launcher.EmulatorLauncher(
          **emulator_launcher_args)
      self._get_emulator_stub = self._launcher.get_emulator_stub
    else:
      self._get_emulator_stub = functools.partial(
          emulator_launcher.EmulatorLauncher.create_emulator_stub,
          grpc_port=self._grpc_port)
    # ZDY_COMMENT: to create an emulator stub is to establish a gRPC connection to the emulator daemon

  def adb_device_name(self) -> str:
    return 'emulator-%s' % (self._adb_port - 1)

  def create_adb_controller(self):
    """Returns an ADB controller which can communicate with this simulator."""
    return adb_controller.AdbController(
        device_name=self.adb_device_name(),
        **self._adb_controller_args)

  # zdy
  #def get_emulator_stub(self):
    #return self._emulator_stub
  @property
  def image_format(self):
    return self._image_format

  def _restart_impl(self) -> None:
    if not self._existing_emulator_provided:
      self._launcher.restart()

  def _launch_impl(self) -> None:
    try:
      if not self._existing_emulator_provided:
        self._launcher.launch()
    except errors.SimulatorCrashError:
      # If the simulator crashes on the initial launch, we try to restart once.
      self.restart()

  def _post_launch_setup(self):
    super()._post_launch_setup()
    if self._adb_root:
      self._adb_controller.get_root()
      logging.info("Inited root adb daemon...")
    if self._frida_server:
      self._adb_controller.init_frida_server(self._frida_server)
      logging.info("Inited frida server...")
    self._emulator_stub = self._get_emulator_stub()
    self._image_format = emulator_controller_pb2.ImageFormat(
        format=emulator_controller_pb2.ImageFormat.ImgFormat.RGB888,
        height=self._screen_dimensions[0],
        width=self._screen_dimensions[1],
    )

  def send_action(self, action: Dict[str, np.ndarray]) -> None:
    """Sends touch events to the emulator."""
    assert self._emulator_stub, 'Emulator stub has not been initialized yet.'
    touches = []
    for i, finger_action in enumerate(self._split_action(action)):
      x, y, down = self._prepare_action(finger_action)
      touches.append(emulator_controller_pb2.Touch(
          x=x, y=y, pressure=int(down), identifier=i))
    self._emulator_stub.sendTouch(
        emulator_controller_pb2.TouchEvent(touches=touches))

  def _get_observation(self) -> Optional[List[np.ndarray]]:
    """Fetches the latest observation from the emulator."""
    assert self._emulator_stub, 'Emulator stub has not been initialized yet.'
    assert self._image_format, 'ImageFormat has not been initialized yet.'
    image_proto = self._emulator_stub.getScreenshot(self._image_format)
    h, w = image_proto.format.height, image_proto.format.width
    image = np.frombuffer(image_proto.image, dtype='uint8', count=h * w * 3)
    image.shape = (h, w, 3)
    return [image, np.int64(image_proto.timestampUs)]

  def _create_log_stream(self) -> log_stream.LogStream:
    return adb_log_stream.AdbLogStream(
        adb_command_prefix=self._adb_controller.command_prefix(),
        verbose=self._verbose_logs)

  def close(self):
    if hasattr(self, '_channel'):
      self._channel.close()
    if hasattr(self, '_emulator_stub'):
      del self._emulator_stub
    if hasattr(self, '_launcher') and not self._existing_emulator_provided:
      self._launcher.close()
    super().close()



# coding=utf-8
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Extends Android observation with the latest action taken."""

from typing import Dict

from android_env.components import action_type
from android_env.components import utils
from android_env.wrappers import base_wrapper
import dm_env
from dm_env import specs
import numpy as np


class LastActionWrapper(base_wrapper.BaseWrapper):
  """Extends Android observations with information about the last action taken.

  The position of the last action is denoted by a single white pixel (with a
  value of 255) in a channel of all black pixels (with a value of 0).
  As this wrapper makes use of temporarily stored information about the
  last action taken, it is important to apply on the environment side rather
  than the agent side. Recommended not to apply before an ImageRescaleWrapper,
  to avoid distortion of the single pixel denoting the action position.
  """

  def __init__(self,
               env: dm_env.Environment,
               concat_to_pixels: bool = True):
    """Initializes the internal state of this wrapper.

    Args:
      env: the environment to wrap.
      concat_to_pixels: If True, will add a channel to the pixel observation.
        If False, will pass the action as an extra observation.
    """
    super().__init__(env)
    self._concat_to_pixels = concat_to_pixels
    self._screen_dimensions = self._env.observation_spec()['pixels'].shape[:2]

  def _process_timestep(self, timestep: dm_env.TimeStep) -> dm_env.TimeStep:
    observation = timestep.observation.copy()
    processed_observation = self._process_observation(observation)
    return timestep._replace(observation=processed_observation)

  def _process_observation(
      self, observation: Dict[str, np.ndarray]
  ) -> Dict[str, np.ndarray]:
    """Extends observation with last_action data."""
    processed_observation = observation.copy()
    last_action_layer = self._get_last_action_layer(observation['pixels'])
    if self._concat_to_pixels:
      pixels = observation['pixels'].copy()
      processed_pixels = np.dstack((pixels, last_action_layer))
      processed_observation['pixels'] = processed_pixels
    else:
      processed_observation['last_action'] = last_action_layer
    return processed_observation

  def _get_last_action_layer(self, pixels: np.ndarray) -> np.ndarray:
    """Makes sure the rescaling doesn't distort the last_action layer."""

    last_action = self._env.raw_action
    last_action_layer = np.zeros(self._screen_dimensions, dtype=pixels.dtype)

    if ('action_type' in last_action and
        last_action['action_type'] == action_type.ActionType.TOUCH):
      touch_position = last_action['touch_position']
      x, y = utils.touch_position_to_pixel_position(
          touch_position, width_height=self._screen_dimensions[::-1])
      last_action_layer[y, x] = 255

    return last_action_layer

  def reset(self) -> dm_env.TimeStep:
    timestep = self._env.reset()
    return self._process_timestep(timestep)

  def step(self, action) -> dm_env.TimeStep:
    timestep = self._env.step(action)
    return self._process_timestep(timestep)

  def observation_spec(self) -> Dict[str, specs.Array]:
    parent_spec = self._env.observation_spec().copy()
    shape = parent_spec['pixels'].shape
    if self._concat_to_pixels:
      parent_spec['pixels'] = specs.Array(
          shape=(shape[0], shape[1], shape[2] + 1),
          dtype=parent_spec['pixels'].dtype,
          name=parent_spec['pixels'].name)
    else:
      parent_spec.update({
          'last_action':
              specs.Array(
                  shape=(shape[0], shape[1]),
                  dtype=parent_spec['pixels'].dtype,
                  name='last_action')
      })
    return parent_spec



# vim: set tabstop=2 shiftwidth=2:
# coding=utf-8
# Copyright 2023 SJTU X-Lance Lab
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Revised by Danyang Zhang @X-Lance based on
# 
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Android environment implementation."""

from typing import Any, Dict, List, Union
from absl import logging
from android_env.components import coordinator as coordinator_lib
from android_env.components import task_manager as task_manager_lib
from android_env.utils import fix_path
from android_env.proto import task_pb2
from google.protobuf import text_format
import os.path
import dm_env
import numpy as np

class AndroidEnv(dm_env.Environment):
  """An RL environment that interacts with Android apps."""

  def __init__(self, coordinator: coordinator_lib.Coordinator):
    """Initializes the state of this AndroidEnv object."""

    self._coordinator = coordinator
    self._latest_action = {}
    self._latest_observation = {}
    self._latest_extras = {}
    self._latest_instruction = []
    self._reset_next_step = True

    logging.info('Action spec: %s', self.action_spec())
    logging.info('Observation spec: %s', self.observation_spec())
    logging.info('Task extras spec: %s', self.task_extras_spec())

  #  Informational Interfaces {{{ # 
  def action_spec(self) -> Dict[str, dm_env.specs.Array]:
    return self._coordinator.action_spec()

  def observation_spec(self) -> Dict[str, dm_env.specs.Array]:
    return self._coordinator.observation_spec()

  def task_extras_spec(self) -> Dict[str, dm_env.specs.Array]:
    return self._coordinator.task_extras_spec()

  def command(self) -> List[str]:
    """
    return list of str
    """

    return self._coordinator.command()

  def vocabulary(self) -> List[str]:
    """
    return list of str
    """

    return self._coordinator.vocabulary()

  @property
  def nb_tasks(self) -> int:
    return self._coordinator.nb_tasks
  @property
  def task_index(self) -> int:
    return self._coordinator.task_index
  @property
  def task_id(self) -> str:
    return self._coordinator.task_id
  @property
  def task_name(self) -> str:
    return self._coordinator.task_name
  @property
  def task_description(self) -> str:
    return self._coordinator.task_description
  #  }}} Informational Interfaces # 

  @property
  def raw_action(self):
    return self._latest_action

  @property
  def raw_observation(self):
    return self._latest_observation

  def android_logs(self) -> Dict[str, Any]:
    return self._coordinator.get_logs()

  def add_task(self, task_path: str):
    #  method `add_task` {{{ # 
    task = task_pb2.Task()
    with open(task_path, 'r') as f:
      text_format.Parse(f.read(), task)

    fix_path(task, os.path.dirname(task_path))
    #for st in task.setup_steps:
      #if st.HasField("adb_call") and\
          #st.adb_call.HasField("install_apk"):
        #apk_path = st.adb_call.install_apk.filesystem.path
        #if not os.path.isabs(apk_path):
          #st.adb_call.install_apk.filesystem.path =\
              #os.path.normpath(
              #os.path.join(os.path.dirname(task_path),
                #apk_path))

    task_manager = task_manager_lib.TaskManager(task)
    self._coordinator.add_task_manager(task_manager)
    #  }}} method `add_task` # 

  def switch_task(self, index: int) -> dm_env.TimeStep:
    #  method `change_task` {{{ # 
    logging.info('Changing Task to {:d}...'.format(index))

    # Change the task and reset state of the environment.
    self._coordinator.switch_task_manager(index)

    # Execute selected action (None when resetting).
    obs, _, extras, instructions, _ = self._coordinator.execute_action(action=None)

    # Process relevant information.
    if obs is not None:
      self._latest_observation = obs.copy()
    self._latest_extras = extras.copy()
    self._latest_instruction = instructions.copy()
    self._latest_action = {}
    self._reset_next_step = False

    logging.info('Done Changing Task.')
    logging.info('************* NEW EPISODE *************')

    return dm_env.TimeStep(
        step_type=dm_env.StepType.FIRST,
        observation=self._latest_observation,
        reward=0.0,
        discount=0.0)
    #  }}} method `change_task` # 

  def reset(self) -> dm_env.TimeStep:
    """Resets the environment for a new RL episode."""

    logging.info('Resetting AndroidEnv...')

    # Reset state of the environment.
    self._coordinator.reset_environment_state()

    # Execute selected action (None when resetting).
    obs, _, extras, instructions, _ = self._coordinator.execute_action(action=None)

    # Process relevant information.
    if obs is not None:
      self._latest_observation = obs.copy()
    self._latest_extras = extras.copy()
    self._latest_instruction = instructions.copy()
    self._latest_action = {}
    self._reset_next_step = False

    logging.info('Done resetting AndroidEnv.')
    logging.info('************* NEW EPISODE *************')

    return dm_env.TimeStep(
        step_type=dm_env.StepType.FIRST,
        observation=self._latest_observation,
        reward=0.0,
        discount=0.0)

  def step(self, action: Dict[str, np.ndarray]) -> dm_env.TimeStep:
    """Takes a step in the environment."""

    # Check if it's time to reset the episode.
    if self._reset_next_step:
      return self.reset()

    # Execute selected action.
    obs, reward, extras, instructions, episode_end = self._coordinator.execute_action(action)

    # Process relevant information.
    if obs is not None:
      self._latest_observation = obs.copy()
    self._latest_extras = extras.copy()
    self._latest_instruction = instructions.copy()
    self._latest_action = action.copy()
    self._reset_next_step = episode_end

    # Return timestep with reward and observation just computed.
    if episode_end:
      return dm_env.termination(
          observation=self._latest_observation, reward=reward)
    else:
      return dm_env.transition(
          observation=self._latest_observation, reward=reward, discount=0.0)

  def task_extras(self, latest_only: bool = True) -> Dict[str, np.ndarray]:
    """Returns latest task extras."""

    task_extras = {}
    for key, spec in self.task_extras_spec().items():
      if key in self._latest_extras:
        extra_values = self._latest_extras[key].astype(spec.dtype)
        for extra in extra_values:
          spec.validate(extra)
        if latest_only:
          task_extras[key] = extra_values[-1] if len(extra_values)>0\
              else np.zeros(shape=spec.shape, dtype=spec.dtype)
        else:
          task_extras[key] = extra_values
    return task_extras

  def task_instructions(self, latest_only: bool = False) -> Union[str, List[str]]:
    """
    latest_only - bool

    return str or list of str
    """

    if latest_only:
      return self._latest_instruction[-1] if len(self._latest_instruction)>0 else ""
    else:
      return self._latest_instruction.copy()

  def close(self) -> None:
    """Cleans up running processes, threads and local files."""
    logging.info('Cleaning up AndroidEnv...')
    if hasattr(self, '_coordinator'):
      self._coordinator.close()
    logging.info('Done cleaning up AndroidEnv.')

  def __del__(self) -> None:
    self.close()



# coding=utf-8
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for android_env.components.coordinator."""

import time

from absl.testing import absltest
from android_env.components import action_type
from android_env.components import coordinator
from android_env.components import errors
from android_env.components import task_manager
from android_env.components.simulators.emulator import emulator_simulator
import mock
import numpy as np


class CoordinatorTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.addCleanup(mock.patch.stopall)  # Disable previous patches.

    self._simulator = mock.create_autospec(emulator_simulator.EmulatorSimulator)
    self._task_manager = mock.create_autospec(task_manager.TaskManager)

    self._coordinator = coordinator.Coordinator(
        simulator=self._simulator,
        task_manager=self._task_manager,
        step_timeout_sec=2,
        max_steps_per_sec=60,
        periodic_restart_time_min=0)

  def test_restart_simulator(self):
    self._coordinator._restart_simulator()

  def test_reset(self):
    self._coordinator.reset_environment_state()

  def test_lift_all_fingers(self):
    self._simulator.num_fingers.return_value = 3
    self._coordinator.reset_environment_state()
    expected_lift_action = {
        'action_type': np.array(action_type.ActionType.LIFT),
        'touch_position': np.array([0, 0]),
        'action_type_2': np.array(action_type.ActionType.LIFT),
        'touch_position_2': np.array([0, 0]),
        'action_type_3': np.array(action_type.ActionType.LIFT),
        'touch_position_3': np.array([0, 0]),
    }
    actual_lift_action = self._simulator.send_action.call_args[0][0]
    for key, value in expected_lift_action.items():
      np.testing.assert_array_equal(actual_lift_action[key], value)

  def test_process_action(self):
    self._simulator.num_fingers.return_value = 1
    self._simulator.get_observation.return_value = {'observation': 0}
    self._task_manager.get_current_reward.return_value = 10.0
    self._task_manager.get_current_extras.return_value = {'extra': [0.0]}
    self._task_manager.check_if_episode_ended.return_value = False
    obs, reward, task_extras, episode_end = self._coordinator.execute_action(
        action={'action_type': np.array(action_type.ActionType.LIFT)})
    self.assertDictEqual(obs, {'observation': 0})
    self.assertEqual(reward, 10.0)
    self.assertEqual(task_extras, {'extra': [0.0]})
    self.assertFalse(episode_end)

  def test_process_action_error(self):
    self._simulator.num_fingers.return_value = 1
    self._simulator.get_observation.side_effect = errors.ReadObservationError()
    self._task_manager.get_current_reward.return_value = 0.0
    self._task_manager.get_current_extras.return_value = {}
    self._task_manager.check_if_episode_ended.return_value = True
    obs, reward, task_extras, episode_end = self._coordinator.execute_action(
        action={'action_type': np.array(action_type.ActionType.LIFT)})
    self.assertTrue(self._coordinator._should_restart)
    self.assertIsNone(obs)
    self.assertEqual(reward, 0.0)
    self.assertEqual(task_extras, {})
    self.assertTrue(episode_end)

  def test_execute_action_touch(self):
    self._simulator.num_fingers.return_value = 1
    self._simulator.get_observation.return_value = {'observation': 0}
    self._simulator.send_action.return_value = True
    action = {'action_type': np.array(action_type.ActionType.TOUCH)}
    _ = self._coordinator.execute_action(action)
    self._simulator.send_action.assert_called_once_with(action)

  def test_execute_action_repeat(self):
    self._simulator.num_fingers.return_value = 1
    self._simulator.get_observation.return_value = {'observation': 0}
    self._simulator.send_action.return_value = True
    _ = self._coordinator.execute_action(
        {'action_type': np.array(action_type.ActionType.REPEAT)})
    self._simulator.send_action.assert_not_called()

  def test_execute_action_error(self):
    self._simulator.num_fingers.return_value = 1
    self._simulator.get_observation.return_value = {'observation': 0}
    self._simulator.send_action.side_effect = errors.SendActionError
    _ = self._coordinator.execute_action(
        {'action_type': np.array(action_type.ActionType.TOUCH)})
    self.assertTrue(self._coordinator._should_restart)

  def test_check_timeout_false(self):
    self._coordinator._latest_observation_time = time.time()
    timeout = self._coordinator._check_timeout()
    self.assertFalse(timeout)

  def test_check_timeout_true(self):
    self._coordinator._latest_observation_time = time.time()
    time.sleep(3)
    timeout = self._coordinator._check_timeout()
    self.assertTrue(timeout)

  def test_max_restarts_adb_error(self):
    # The method was called once at init.
    init_fn_call = self._simulator.create_adb_controller.call_count
    self._simulator.create_adb_controller.side_effect = (
        errors.AdbControllerError)
    self.assertRaises(errors.TooManyRestartsError,
                      self._coordinator._restart_simulator)
    # The method was called three more times when attempting to restart.
    self.assertEqual(init_fn_call + 3,
                     self._simulator.create_adb_controller.call_count)

  def test_max_restarts_setup_steps(self):
    init_fn_call = self._task_manager.setup_task.call_count
    self._task_manager.setup_task.side_effect = errors.StepCommandError
    self.assertRaises(errors.TooManyRestartsError,
                      self._coordinator._restart_simulator)
    # The method was called three more times when attempting to restart.
    self.assertEqual(init_fn_call + 3,
                     self._task_manager.setup_task.call_count)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 SJTU X-Lance Lab
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Created by Danyang Zhang @X-Lance.
"""

import abc
import torch
from typing import List, Tuple

class TextModel(abc.ABC):
    def text_detector( self
                     , screen: torch.Tensor
                     , bboxes: List[torch.Tensor]
                     ) -> List[List[str]]:
        """
        Args:
            screen: torch.Tensor of float32 with shape (3, height, width)
            bboxes: list with length nb_bboxes of torch.Tensor of float32 with shape (1, 4)

        Returns:
            list with length nb_bboxes of list of str
        """

        return [[] for _ in bboxes]
    def text_recognizer( self
                       , screen: torch.Tensor
                       , bboxes: List[torch.Tensor]
                       ) -> List[str]:
        """
        Args:
            screen: torch.Tensor of float32 with shape (3, height, width)
            bboxes: list with length nb_bboxes of torch.Tensor of float32 with shape (1, 4)

        Returns:
            list with length nb_bboxes of str
        """

        return ["" for _ in bboxes]

class IconModel(abc.ABC):
    def icon_detector( self
                     , screen: torch.Tensor
                     , bboxes: List[torch.Tensor]
                     ) -> Tuple[ torch.Tensor
                               , List[List[str]]
                               ]:
        """
        Args:
            screen: torch.Tensor of float32 with shape (3, height, width)
            bboxes: list with length nb_bboxes of torch.Tensor of float32 with shape (1, 4)

        Returns:
            - torch.Tensor of float with shape (nb_bboxes, nb_candidates, 4)
            - list with length nb_bboxes of list with length nb_candidates of
              str as the icon classes
        """

        return torch.stack(bboxes)\
             , [[""] for _ in bboxes]
    def icon_recognizer( self
                       , screen: torch.Tensor
                       , bboxes: List[torch.Tensor]
                        ) -> List[str]:
        """
        Args:
            screen: torch.Tensor of float32 with shape (3, height, width)
            bboxes: list with length nb_bboxes of torch.Tensor of float32 with shape (1, 4)

        Returns:
            list with length nb_bboxes of str
        """

        return ["" for _ in bboxes]
    def icon_matcher( self
                    , screen: torch.Tensor
                    , targets: List[torch.Tensor]
                    , bboxes: torch.Tensor
                    ) -> List[List[bool]]:
        """
        Args:
            screen: torch.Tensor of float32 with shape (3, height, width)
            targets: list with length nb_bboxes of torch.Tensor of float32 with
              shape (3, height', width')
            bboxes: torch.Tensor of float32 with shape (nb_bboxes, nb_candidates, 4)

        Returns:
            list with length nb_bboxes of list with length nb_candidates of bool
        """

        return [[False for _ in bb] for bb in bboxes]



# coding=utf-8
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Class for a stream of logs output by a locally running emulator."""

import subprocess
from typing import List

from android_env.components import log_stream


_LOGCAT_COMMAND = ['logcat', '-v', 'epoch']


class AdbLogStream(log_stream.LogStream):
  """Manages adb logcat process for a locally running emulator."""

  def __init__(self, adb_command_prefix: List[str], *args, **kwargs):
    super().__init__(*args, **kwargs)
    self._adb_command_prefix = adb_command_prefix

  def _get_stream_output(self):
    cmd = self._adb_command_prefix + _LOGCAT_COMMAND + self._filters
    self._adb_subprocess = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
        universal_newlines=True)
    return self._adb_subprocess.stdout

  def stop_stream(self):
    self._adb_subprocess.kill()



# coding=utf-8
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for android_env.components.base_simulator."""

from typing import List, Optional

from absl.testing import absltest
from android_env.components import action_type
from android_env.components import adb_controller
from android_env.components.simulators import base_simulator
import mock
import numpy as np


class FakeSimulator(base_simulator.BaseSimulator):
  """A simulator that spits injected data."""

  def __init__(self, injected_adb_controller, **kwargs):
    super().__init__(**kwargs)
    self._injected_adb_controller = injected_adb_controller
    self._adb_controller = self.create_adb_controller()

  def adb_device_name(self) -> str:
    return 'FakeSimulator'

  def create_adb_controller(self) -> adb_controller.AdbController:
    return self._injected_adb_controller

  def send_action(self, action) -> None:
    pass

  def _restart_impl(self) -> None:
    pass

  def _launch_impl(self) -> None:
    pass

  def _get_observation(self) -> Optional[List[np.ndarray]]:
    return [np.ones(shape=(640, 480, 3)), self._timestamp]

  def _create_log_stream(self):
    pass

  def set_timestamp(self, timestamp: int):
    self._timestamp = timestamp


class BaseSimulatorTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self._adb_controller = self.enter_context(
        mock.patch.object(adb_controller, 'AdbController', autospec=True))
    self._simulator = FakeSimulator(
        injected_adb_controller=self._adb_controller)

  def test_device_name(self):
    self.assertNotEmpty(self._simulator.adb_device_name())

  def test_launch(self):
    # Before a successful launch(), screen_dimensions() should raise.
    self.assertRaises(AssertionError, self._simulator.screen_dimensions)
    # The simulator should launch and not crash.
    self._adb_controller.get_screen_dimensions.return_value = (640, 480)
    self._simulator.launch()
    # After a successful launch(), screen_dimensions() should return something.
    np.testing.assert_equal(self._simulator.screen_dimensions(), [640, 480])

  def test_launch_close(self):
    # The simulator should launch and not crash.
    self._adb_controller.get_screen_dimensions.return_value = (640, 480)
    self._simulator.launch()
    # Closing the simulator should also not crash.
    self._simulator.close()

  def test_get_observation(self):
    # The simulator should launch and not crash.
    self._adb_controller.get_screen_dimensions.return_value = (640, 480)
    self._simulator.launch()

    self._simulator.set_timestamp(123456)
    observation = self._simulator.get_observation()
    # Even though the FakeSimulator returns a 640x480x3 image, BaseSimulator
    # should add an extra layer for the last action.
    np.testing.assert_equal(observation['pixels'].shape, [640, 480, 3])
    # Because there was only a single step, the timestamp delta should be just
    # the timestamp returned by FakeSimulator.
    self.assertEqual(observation['timedelta'], 123456)

    # The orientation format should be a 4-dimension one-hot.
    np.testing.assert_equal(observation['orientation'].shape, [4])
    # Initially the orientation should not be set so we expect [0, 0, 0, 0].
    self.assertEqual(np.sum(observation['orientation']), 0)

    # After updating the device orientation, we do expect to get an actual
    # one-hot.
    self._adb_controller.get_orientation.return_value = '3'
    self._simulator.update_device_orientation()

    self._simulator.set_timestamp(123459)
    updated_observation = self._simulator.get_observation()
    np.testing.assert_equal(updated_observation['orientation'], [0, 0, 0, 1])
    # After setting the initial timestamp, we expect the timestamp delta to be
    # the difference between the current timestamp and the initial timestamp:
    # 123459 - 123456 = 3
    self.assertEqual(updated_observation['timedelta'], 3)

  def test_prepare_action(self):
    # The simulator should launch and not crash.
    self._adb_controller.get_screen_dimensions.return_value = (480, 640)
    self._simulator.launch()

    self._simulator.set_timestamp(123456)
    self.assertEqual(
        self._simulator._prepare_action({
            'action_type': np.array(action_type.ActionType.LIFT),
            'touch_position': [0.0, 0.0]
        }), (0, 0, False))
    self.assertEqual(
        self._simulator._prepare_action({
            'action_type': np.array(action_type.ActionType.LIFT),
            'touch_position': [0.5, 0.2]
        }), (320, 96, False))
    self.assertEqual(
        self._simulator._prepare_action({
            'action_type': np.array(action_type.ActionType.TOUCH),
            'touch_position': [0.0, 0.0]
        }), (0, 0, True))
    self.assertEqual(
        self._simulator._prepare_action({
            'action_type': np.array(action_type.ActionType.TOUCH),
            'touch_position': [0.5, 0.2]
        }), (320, 96, True))
    self.assertEqual(
        self._simulator._prepare_action({
            'action_type': np.array(action_type.ActionType.TOUCH),
            'touch_position': [1.0, 1.0]
        }), (639, 479, True))


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for loader."""

import builtins
import os
from absl.testing import absltest
from android_env import environment
from android_env import loader
from android_env.components import coordinator as coordinator_lib
from android_env.components import task_manager as task_manager_lib
from android_env.components.simulators.emulator import emulator_simulator
from android_env.proto import task_pb2
import mock


class LoaderTest(absltest.TestCase):

  @mock.patch.object(task_manager_lib, 'TaskManager', autospec=True)
  @mock.patch.object(emulator_simulator, 'EmulatorSimulator', autospec=True)
  @mock.patch.object(coordinator_lib, 'Coordinator', autospec=True)
  @mock.patch.object(builtins, 'open', autospec=True)
  def test_load(self, mock_open, coordinator, simulator, task_manager):

    mock_open.return_value.__enter__ = mock_open
    mock_open.return_value.read.return_value = ''

    env = loader.load(
        task_path='some/path/',
        avd_name='my_avd',
        android_avd_home='~/.android/avd',
        android_sdk_root='~/Android/Sdk',
        emulator_path='~/Android/Sdk/emulator/emulator',
        adb_path='~/Android/Sdk/platform-tools/adb',
        run_headless=False,
    )

    self.assertIsInstance(env, environment.AndroidEnv)
    simulator.assert_called_with(
        adb_controller_args=dict(
            adb_path=os.path.expanduser('~/Android/Sdk/platform-tools/adb'),
            adb_server_port=5037,
            prompt_regex=r'\w*:\/ \$',
        ),
        emulator_launcher_args=dict(
            avd_name='my_avd',
            android_avd_home=os.path.expanduser('~/.android/avd'),
            android_sdk_root=os.path.expanduser('~/Android/Sdk'),
            emulator_path=os.path.expanduser('~/Android/Sdk/emulator/emulator'),
            run_headless=False,
            gpu_mode='swiftshader_indirect'),
    )
    coordinator.assert_called_with(
        simulator.return_value,
        task_manager.return_value,
    )

  @mock.patch.object(task_manager_lib, 'TaskManager', autospec=True)
  @mock.patch.object(emulator_simulator, 'EmulatorSimulator', autospec=True)
  @mock.patch.object(coordinator_lib, 'Coordinator', autospec=True)
  @mock.patch.object(builtins, 'open', autospec=True)
  def test_task(self, mock_open, coordinator, simulator, task_manager):

    del coordinator, simulator
    mock_open.return_value.__enter__ = mock_open
    mock_open.return_value.read.return_value = r'''
id: "fake_task"
name: "Fake Task"
description: "Task for testing loader."
max_duration_sec: 0
'''

    env = loader.load(
        task_path='some/path/',
        avd_name='my_avd',
    )

    expected_task = task_pb2.Task()
    expected_task.id = 'fake_task'
    expected_task.name = 'Fake Task'
    expected_task.description = 'Task for testing loader.'
    expected_task.max_duration_sec = 0

    task_manager.assert_called_with(expected_task)
    assert isinstance(env, environment.AndroidEnv)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2023 SJTU X-Lance Lab
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Revised by Danyang Zhang @X-Lance based on
# 
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Init file for android_env.wrappers package."""

from android_env.wrappers import discrete_action_wrapper
from android_env.wrappers import flat_interface_wrapper
from android_env.wrappers import float_pixels_wrapper
from android_env.wrappers import gym_wrapper
from android_env.wrappers import image_rescale_wrapper
from android_env.wrappers import last_action_wrapper
from android_env.wrappers import vh_io_wrapper

DiscreteActionWrapper = discrete_action_wrapper.DiscreteActionWrapper
FlatInterfaceWrapper = flat_interface_wrapper.FlatInterfaceWrapper
FloatPixelsWrapper = float_pixels_wrapper.FloatPixelsWrapper
GymInterfaceWrapper = gym_wrapper.GymInterfaceWrapper
ImageRescaleWrapper = image_rescale_wrapper.ImageRescaleWrapper
LastActionWrapper = last_action_wrapper.LastActionWrapper
VhIoWrapper = vh_io_wrapper.VhIoWrapper



# coding=utf-8
# vim: set tabstop=2 shiftwidth=2:
# Copyright 2023 SJTU X-Lance Lab
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Revised by Danyang Zhang @X-Lance based on
# 
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A ThreadFunction that runs and parses adb dumpsys."""

import enum
import threading

from absl import logging

from android_env.components import app_screen_checker as screen_checker
from android_env.components import thread_function
from android_env.components import event_listeners

AppScreenChecker = screen_checker.AppScreenChecker
# ZDY_COMMENT: check match according to the view hierarchy path


class DumpsysThread(thread_function.ThreadFunction):
  """A class that executes dumpsys in a separate thread."""

  class Signal(enum.IntEnum):
    """Defines commands we can use to communicate with the dumpsys thread."""
    # To ask the thread to fetch dumpsys from Android.
    FETCH_DUMPSYS = 0
    # The user has left the activity that contains the AndroidEnv task.
    USER_EXITED_ACTIVITY = 1
    # The user exited the view hierarchy that we expect.
    USER_EXITED_VIEW_HIERARCHY = 2
    # App screen checker determined none of the errors above happened.
    OK = 3
    # App screen checker was not queried.
    DID_NOT_CHECK = 4

  def __init__(
      self,
      app_screen_checker: AppScreenChecker,
      check_frequency: int,
      max_failed_current_activity: int,
      lock: threading.Lock, # zdy
      block_input: bool,
      block_output: bool,
      name: str = 'dumpsys',
  ):
    #  method `__init__` {{{ # 
    """Initializes the dumpsys reader thread.

    This loops forever waiting for inputs from the main thread and outputting
    its analyses of the output of ADB dumpsys. These analyses are too expensive
    to be in the critical path of AndroidEnv::step() so we consume them async
    from this separate thread.

    Args:
      app_screen_checker: The class that actually determines if the current
          screen matches the expected screen.
      check_frequency: Integer. We only call dumpsys 1/check_frequency times in
          each iteration of the while loop below.
      max_failed_current_activity: Integer. We try to fetch the current activity
          but sometimes it fails. If it fails more than
          `max_failed_current_activity` consecutive times, we declare that the
          user has exited `expected_activity`.
      lock: Lock for consistency during set flags of events.
      block_input: Whether to block this thread when reading its input queue.
      block_output: Whether to block this thread when writing to its output
        queue.
      name: Name of the thread.
    """

    # zdy
    #self._event_listeners = []
    self._lock = lock

    self._app_screen_checker = app_screen_checker
    self._main_loop_counter = 0
    self._check_frequency = check_frequency
    self._max_failed_activity_extraction = max_failed_current_activity
    self._num_failed_activity_extraction = 0
    super().__init__(
        block_input=block_input, block_output=block_output, name=name)
    #  }}} method `__init__` # 

  def main(self):
    #  method `main` {{{ # 
    v = self._read_value()
    if v != DumpsysThread.Signal.FETCH_DUMPSYS:
      self._write_value(DumpsysThread.Signal.DID_NOT_CHECK)
      return

    # Update and check loop_counter against check_frequency.
    self._main_loop_counter += 1
    if (self._check_frequency <= 0 or
        self._main_loop_counter < self._check_frequency):
      self._write_value(DumpsysThread.Signal.DID_NOT_CHECK)
      return
    self._main_loop_counter = 0

    # The reward signals should be checked anyway, I think.
    # However, it's too slow, so I moved it here.
    self._app_screen_checker.match_events(self._lock)

    outcome = self._app_screen_checker.matches_current_app_screen()

    # We were unable to determine the current activity.
    if outcome == AppScreenChecker.Outcome.FAILED_ACTIVITY_EXTRACTION:
      self._num_failed_activity_extraction += 1
      logging.info('self._num_failed_activity_extraction: %s',
                   self._num_failed_activity_extraction)
      if (self._num_failed_activity_extraction >=
          self._max_failed_activity_extraction):
        logging.error('Maximum number of failed activity extraction reached.')
        self._num_failed_activity_extraction = 0
        self._write_value(DumpsysThread.Signal.USER_EXITED_ACTIVITY)
        return
    else:
      self._num_failed_activity_extraction = 0

    # The current app screen matches all expectations.
    if (outcome == AppScreenChecker.Outcome.SUCCESS or
        outcome == AppScreenChecker.Outcome.EMPTY_EXPECTED_ACTIVITY):
      self._write_value(DumpsysThread.Signal.OK)
      return

    # Player has exited the app. Terminate the episode.
    elif outcome == AppScreenChecker.Outcome.UNEXPECTED_ACTIVITY:
      self._write_value(DumpsysThread.Signal.USER_EXITED_ACTIVITY)
      return

    # Player has exited the main game. Terminate the episode.
    elif outcome == AppScreenChecker.Outcome.UNEXPECTED_VIEW_HIERARCHY:
      self._write_value(DumpsysThread.Signal.USER_EXITED_VIEW_HIERARCHY)
      return
    #  }}} method `main` # 



# coding=utf-8
# Copyright 2023 SJTU X-Lance Lab
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Revised by Danyang Zhang @X-Lance based on
# 
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wraps the AndroidEnv environment to provide tap actions of a given duration."""

from typing import Dict, Sequence

from android_env.components import action_type
from android_env.wrappers import base_wrapper
import dm_env
import numpy as np


ActionType = action_type.ActionType


class TapActionWrapper(base_wrapper.BaseWrapper):
  """AndroidEnv with tap actions."""

  def __init__(self,
               env: dm_env.Environment,
               num_frames: int = 5,
               touch_only: bool = False):
    super().__init__(env)
    assert 'action_type' in env.action_spec()
    self._touch_only = touch_only
    self._num_frames = num_frames
    self._env_steps = 0

  def android_logs(self):
    """Returns a dictionary of metrics logged by the environment."""
    logs = self._env.android_logs()
    logs.update({'env_steps': self._env_steps})
    return logs

  def _process_action(
      self, action: Dict[str, np.ndarray]
  ) -> Sequence[Dict[str, np.ndarray]]:

    if self._touch_only:
      assert action['action_type'] == 0
      touch_action = action.copy()
      touch_action['action_type'] = np.array(ActionType.TOUCH).astype(
          self.action_spec()['action_type'].dtype)
      actions = [touch_action] * self._num_frames
      lift_action = action.copy()
      lift_action['action_type'] = np.array(ActionType.LIFT).astype(
          self.action_spec()['action_type'].dtype)
      actions.append(lift_action)

    else:
      if action['action_type'] == ActionType.TOUCH:
        actions = [action] * self._num_frames
        lift_action = action.copy()
        lift_action['action_type'] = np.array(ActionType.LIFT).astype(
            self.action_spec()['action_type'].dtype)
        actions.append(lift_action)
      else:
        actions = [action]

    return actions

  def step(self, action: Dict[str, np.ndarray]) -> dm_env.TimeStep:
    """Takes a step in the environment."""
    actions = self._process_action(action)
    self._env_steps += len(actions)
    total_reward = 0
    for idx in range(len(actions)):
      step_type, reward, discount, observation = self._env.step(actions[idx])
      if reward:
        total_reward += reward
      if step_type == dm_env.StepType.LAST:
        return dm_env.TimeStep(
            step_type=step_type,
            reward=total_reward,
            discount=discount,
            observation=observation)
    return dm_env.TimeStep(
        step_type=step_type,
        reward=total_reward,
        discount=discount,
        observation=observation)

  def action_spec(self) -> Dict[str, dm_env.specs.Array]:
    if self._touch_only:
      return {
          'action_type':
              dm_env.specs.DiscreteArray(num_values=1, name='action_type'),
          'touch_position':
              self._env.action_spec()['touch_position'],
      }
    else:
      return self._env.action_spec()




# coding=utf-8
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for android_env.components.dumpsys_thread."""

from absl.testing import absltest
from android_env.components import app_screen_checker as screen_checker
from android_env.components import dumpsys_thread

import mock


class DumpsysThreadTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self._dumpsys_thread = dumpsys_thread.DumpsysThread(
        app_screen_checker=mock.create_autospec(
            screen_checker.AppScreenChecker),
        check_frequency=1,
        max_failed_current_activity=5,
        block_input=True,
        block_output=True)

  def test_unexpected_activity(self):
    outcome = screen_checker.AppScreenChecker.Outcome.UNEXPECTED_ACTIVITY
    self._dumpsys_thread._app_screen_checker.matches_current_app_screen.return_value = outcome
    self._dumpsys_thread.write(
        dumpsys_thread.DumpsysThread.Signal.FETCH_DUMPSYS)
    v = self._dumpsys_thread.read(block=True)
    expected = dumpsys_thread.DumpsysThread.Signal.USER_EXITED_ACTIVITY
    self.assertEqual(expected, v)

  def test_unexpected_view_hierarchy(self):
    outcome = screen_checker.AppScreenChecker.Outcome.UNEXPECTED_VIEW_HIERARCHY
    self._dumpsys_thread._app_screen_checker.matches_current_app_screen.return_value = outcome
    self._dumpsys_thread.write(
        dumpsys_thread.DumpsysThread.Signal.FETCH_DUMPSYS)
    v = self._dumpsys_thread.read(block=True)
    expected = dumpsys_thread.DumpsysThread.Signal.USER_EXITED_VIEW_HIERARCHY
    self.assertEqual(expected, v)

  def test_success(self):
    outcome = screen_checker.AppScreenChecker.Outcome.SUCCESS
    self._dumpsys_thread._app_screen_checker.matches_current_app_screen.return_value = outcome
    self._dumpsys_thread.write(
        dumpsys_thread.DumpsysThread.Signal.FETCH_DUMPSYS)
    v = self._dumpsys_thread.read(block=True)
    expected = dumpsys_thread.DumpsysThread.Signal.OK
    self.assertEqual(expected, v)

  def test_not_requested(self):
    self._dumpsys_thread.write('wrong_signal')
    v = self._dumpsys_thread.read(block=True)
    expected = dumpsys_thread.DumpsysThread.Signal.DID_NOT_CHECK
    self.assertEqual(expected, v)

  def test_skipped(self):
    outcome = screen_checker.AppScreenChecker.Outcome.SUCCESS
    self._dumpsys_thread._app_screen_checker.matches_current_app_screen.return_value = outcome
    self._dumpsys_thread._check_frequency = 5
    self._dumpsys_thread._main_loop_counter = 0

    for _ in range(4):
      self._dumpsys_thread.write(
          dumpsys_thread.DumpsysThread.Signal.FETCH_DUMPSYS)
      v = self._dumpsys_thread.read(block=True)
      expected = dumpsys_thread.DumpsysThread.Signal.DID_NOT_CHECK
      self.assertEqual(expected, v)

    self._dumpsys_thread.write(
        dumpsys_thread.DumpsysThread.Signal.FETCH_DUMPSYS)
    v = self._dumpsys_thread.read(block=True)
    expected = dumpsys_thread.DumpsysThread.Signal.OK
    self.assertEqual(expected, v)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2023 SJTU X-Lance Lab
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Revised by Danyang Zhang @X-Lance based on
# 
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wraps the AndroidEnv environment to make its interface flat."""

from typing import Union, Dict, Any

from android_env.wrappers import base_wrapper
import dm_env
from dm_env import specs
import numpy as np

RGB_CHANNELS = (0, 1, 2)


def _extract_screen_pixels(obs: np.ndarray):
  """Get only screen pixels by removing previous action layer."""
  is_grayscale_image = obs.shape[-1] == 2
  if is_grayscale_image:
    return np.expand_dims(obs[..., 0], -1)
  return obs[..., RGB_CHANNELS]


def _get_no_action_observation_spec(obs_spec: specs.BoundedArray):
  """Create an observation spec without the action layer."""
  shape = np.array(obs_spec.shape)
  shape[2] -= 1
  minimum = obs_spec.minimum
  maximum = obs_spec.maximum
  is_scalar = lambda x: np.isscalar(x) or np.ndim(x) == 0
  if not is_scalar(minimum):
    minimum = _extract_screen_pixels(minimum)
  if not is_scalar(maximum):
    maximum = _extract_screen_pixels(maximum)
  return obs_spec.replace(shape=shape, minimum=minimum, maximum=maximum)


class FlatInterfaceWrapper(base_wrapper.BaseWrapper):
  """Simple interface for AndroidEnv.

  Removes the structure from observations and actions, keeping only the pixel
  observations. Also exposes action as an int32 scalar, making it easier to use
  with conventional discrete agents. This wrapper expects a discretized action
  space.
  """

  def __init__(self,
               env: dm_env.Environment,
               flat_actions: bool = True,
               flat_observations: bool = True,
               keep_action_layer: bool = True):
    super().__init__(env)
    self._flat_actions = flat_actions
    self._flat_observations = flat_observations
    self._keep_action_layer = keep_action_layer
    self._action_name = list(self._env.action_spec())[0]
    self._assert_base_env()

  def _assert_base_env(self):
    base_action_spec = self._env.action_spec()
    assert len(base_action_spec) == 1, self._env.action_spec()
    assert isinstance(base_action_spec, dict)
    assert isinstance(base_action_spec[self._action_name], specs.BoundedArray)

  def _process_action(self, action: Union[int, np.ndarray, Dict[str, Any]]):
    if self._flat_actions:
      return {self._action_name: action}
    else:
      return action

  def _process_timestep(self, timestep: dm_env.TimeStep) -> dm_env.TimeStep:
    if self._flat_observations:
      step_type, reward, discount, observation = timestep
      # Keep only the pixels.
      pixels = observation['pixels']
      pixels = pixels if self._keep_action_layer else _extract_screen_pixels(
          pixels)
      return dm_env.TimeStep(
          step_type=step_type,
          reward=reward,
          discount=discount,
          observation=pixels)
    else:
      return timestep

  #def reset(self) -> dm_env.TimeStep:
    #timestep = self._env.reset()
    #return self._process_timestep(timestep)

  def step(self, action: int) -> dm_env.TimeStep:
    timestep = self._env.step(self._process_action(action))
    return self._process_timestep(timestep)

  def observation_spec(self) -> Union[specs.Array, Dict[str, specs.Array]]:
    if self._flat_observations:
      pixels_spec = self._env.observation_spec()['pixels']
      if not self._keep_action_layer:
        return _get_no_action_observation_spec(pixels_spec)
      return pixels_spec
    else:
      return self._env.observation_spec()

  def action_spec(self) -> Union[specs.Array, Dict[str, specs.Array]]:
    if self._flat_actions:
      return self._env.action_spec()[self._action_name]
    else:
      return self._env.action_spec()
